# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Base class for all Freedombox applications.
"""

import collections
import enum
import inspect
import logging
from typing import ClassVar, TypeAlias

from plinth import cfg
from plinth.diagnostic_check import DiagnosticCheck
from plinth.signals import post_app_loading

from . import clients as clients_module
from . import db

logger = logging.getLogger(__name__)

_list_type: TypeAlias = list


class App:
    """Implement common functionality for an app.

    An app is composed of components which actually performs various tasks. App
    itself delegates tasks for individual components. Applications can show a
    variation in their behavior by choosing which components to have and by
    customizing the components themselves.

    'app_id' property of the app must be a string that is a globally unique ID.
    This is typically also the name of the python module handling the app. So,
    it should be all lower-case English alphabet and digits without any special
    characters.

    'can_be_disabled' is a boolean indicating whether an app can be disabled by
    the user. Enable/disable button for this app will not be shown. Default
    value is True, so the app can be disabled.

    'locked' is a boolean indicating whether the user can perform operations on
    the app. This flag is currently set during backup and restore operations
    but UI changes are currently not implemented.

    'configure_when_disabled' is a boolean indicating whether the app can
    configured while it is disabled. Some apps such those whose configuration
    is stored in a database can't be configured while they are disabled because
    the database server may not be running when the app is disabled.
    """

    app_id: str | None = None

    can_be_disabled: bool = True

    locked: bool = False  # Whether user interaction with the app is allowed.
    # XXX: Lockdown the application UI by implementing a middleware

    configure_when_disabled: bool = True

    _all_apps: ClassVar[collections.OrderedDict[
        str, 'App']] = collections.OrderedDict()

    class SetupState(enum.Enum):
        """Various states of app being setup."""

        NEEDS_SETUP = 'needs-setup'
        NEEDS_UPDATE = 'needs-update'
        UP_TO_DATE = 'up-to-date'

    def __init__(self) -> None:
        """Build the app by adding components.

        App may be built just for the purpose for querying. For example, when
        querying the list of package dependencies of essential apps, an app is
        minimally constructed under a read-only environment for querying from a
        specific component. So, this operation should have no side-effects such
        connecting to signals, running configuration corrections and scheduling
        operations.
        """
        if not self.app_id:
            raise ValueError('Invalid app ID configured')

        self.components: dict[str, Component] = collections.OrderedDict()

        # Add self to global list of apps
        self._all_apps[self.app_id] = self

    def post_init(self):
        """Perform post initialization operations.

        Additional initialization operations such as connecting to signals,
        running configuration corrections and scheduling operations should be
        done in this method rather than in __init__().
        """

    @classmethod
    def get(cls, app_id):
        """Return an app with given ID."""
        return cls._all_apps[app_id]

    @classmethod
    def list(cls):
        """Return a list of all apps."""
        return cls._all_apps.values()

    def add(self, component):
        """Add a component to an app."""
        component.app_id = self.app_id
        self.components[component.component_id] = component
        return self

    def remove(self, component_id):
        """Remove a component from the app."""
        component = self.components[component_id]
        component.app_id = None
        del self.components[component_id]
        return component

    def get_component(self, component_id):
        """Return a component given the component's ID."""
        return self.components[component_id]

    def get_components_of_type(self, component_type):
        """Return all components of a given type."""
        for component in self.components.values():
            if isinstance(component, component_type):
                yield component

    @property
    def info(self):
        """Return the information component of the app.

        It is mandatory to have one :class:`~plinth.app.Info` component in
        every app to provide basic information about the app. Trying to access
        this property without having the Info component will result in a
        KeyError exception being raised. The lookup for the Info component is
        performed using the auto-generated component_id assigned to the Info
        component based on the app_id.

        """
        return self.get_component(self.app_id + '-info')

    def is_available(self) -> bool:
        """Return whether the app is available to install."""
        for component in self.components.values():
            if not component.is_available():
                return False

        return True

    def setup(self, old_version):
        """Install and configure the app and its components."""
        for component in self.components.values():
            component.setup(old_version=old_version)

    def uninstall(self):
        """De-configure and uninstall the app."""
        # Remove components in the reverse order so that dependencies among
        # components is properly satisfied.
        for component in reversed(self.components.values()):
            component.uninstall()

    def get_setup_state(self) -> SetupState:
        """Return whether the app is not setup or needs upgrade."""
        current_version = self.get_setup_version()
        if current_version and self.info.version <= current_version:
            return self.SetupState.UP_TO_DATE

        if not current_version:
            return self.SetupState.NEEDS_SETUP

        return self.SetupState.NEEDS_UPDATE

    def get_setup_version(self) -> int:
        """Return the setup version of the app."""
        # XXX: Optimize version gets
        from . import models

        try:
            with db.lock:
                app_entry = models.Module.objects.get(pk=self.app_id)
                return app_entry.setup_version
        except models.Module.DoesNotExist:
            return 0

    def needs_setup(self) -> bool:
        """Return whether the app needs to be setup.

        A simple shortcut for get_setup_state() == NEEDS_SETUP
        """
        return self.get_setup_state() == self.SetupState.NEEDS_SETUP

    def set_setup_version(self, version: int) -> None:
        """Set the app's setup version."""
        from . import models

        with db.lock:
            models.Module.objects.update_or_create(
                pk=self.app_id, defaults={'setup_version': version})

    def enable(self):
        """Enable all the components of the app."""
        for component in self.components.values():
            component.enable()

    def disable(self):
        """Enable all the components of the app."""
        for component in reversed(self.components.values()):
            component.disable()

    def is_enabled(self):
        """Return whether all the leader components are enabled.

        Return True when there are no leader components.
        """
        return all((component.is_enabled()
                    for component in self.components.values()
                    if component.is_leader))

    def set_enabled(self, enabled):
        """Update the status of all follower components.

        Do not query or update the status of the leader components.
        """
        for component in self.components.values():
            if not component.is_leader:
                component.set_enabled(enabled)

    def diagnose(self) -> _list_type[DiagnosticCheck]:
        """Run diagnostics and return results.

        Return value must be a list of results. Each result is a
        :class:`~plinth.diagnostic_check.DiagnosticCheck` with a
        unique check_id, a user visible description of the test, the result,
        test parameters, and the component ID. The test result is a string
        enumeration from 'failed', 'passed', 'error', 'warning' and 'not_done'.

        Results are typically collected by diagnosing each component of the app
        and then supplementing the results with any app level diagnostic tests.

        Also see :meth:`.has_diagnostics`.
        """
        results = []
        for component in self.components.values():
            results.extend(component.diagnose())

        return results

    def has_diagnostics(self):
        """Return whether at least one diagnostic test is implemented.

        If this method returns True, a button or menu item is shown to the
        user to run diagnostics on this app. When the action is selected by the
        user, the :meth:`.diagnose` method is called and the results are
        presented to the user. Additionally collection of diagnostic results of
        all apps can be obtained by the user from the Diagnostics module in
        System section.

        If a component of this app implements a diagnostic test, this method
        returns True.

        Further, if a subclass of App overrides the :meth:`.diagnose` method,
        it is assumed that it is for implementing diagnostic tests and this
        method returns True for such an app. Override this method if this
        default behavior does not fit the needs.
        """
        # App implements some diagnostics
        if self.__class__.diagnose is not App.diagnose:
            return True

        # Any of the components implement diagnostics
        for component in self.components.values():
            if component.has_diagnostics():
                return True

        return False

    def repair(self, failed_checks: _list_type) -> bool:
        """Try to fix failed diagnostics.

        The default implementation asks relevant components to repair, and then
        requests re-run setup for the app.

        failed_checks is a list of DiagnosticChecks that had failed or resulted
        in a warning. The list will be split up by component_id, and passed to
        the appropriate components. Remaining failed diagnostics do not have a
        related component, and should be handled by the app.

        Returns whether the app setup should be re-run.
        """
        should_rerun_setup = False

        # Group the failed_checks by component
        components_failed_checks = collections.defaultdict(list)
        for failed_check in failed_checks:
            if failed_check.component_id:
                components_failed_checks[failed_check.component_id].append(
                    failed_check)
            else:
                # There is a failed check with no related component.
                should_rerun_setup = True

        # Repair each component that has failed checks
        for component_id, component in self.components.items():
            if components_failed_checks[component_id]:
                result = component.repair(
                    components_failed_checks[component_id])
                should_rerun_setup = should_rerun_setup or result

        return should_rerun_setup

    def has_logs(self) -> bool:
        """Return if any of the components emit logs.

        This is typically true if the apps has daemons or containers.
        """
        from plinth import log
        for component in self.components.values():
            if isinstance(component, log.LogEmitter):
                return True

        return False

    def get_logs(self) -> _list_type[dict[str, str]]:
        """Return the logs in a dictionary format."""
        from plinth import log
        logs: list[dict[str, str]] = []
        for component in self.components.values():
            if isinstance(component, log.LogEmitter):
                logs.append(component.get_logs())

        return logs


class Component:
    """Interface for an app component.

    `app_id` is a string which is set to the value of the application's app_id
    to which this component belongs. It is set when the component is added to
    an app. When the component is removed from an app, it set to None.

    """

    is_leader = False

    def __init__(self, component_id):
        """Initialize the component."""
        if not component_id:
            raise ValueError('Invalid component ID')

        self.component_id = component_id
        self.app_id = None

    @property
    def app(self):
        """Return the app this component is part of.

        Raises KeyError if this component is not part of any app.

        """
        return App.get(self.app_id)

    def is_available(self) -> bool:
        """Return whether the app is available to install."""
        return True

    def setup(self, old_version):
        """Run operations to install and configure the component."""

    def uninstall(self):
        """De-configure and uninstall the component."""

    def enable(self):
        """Run operations to enable the component."""

    def disable(self):
        """Run operations to disable the component."""

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return results.

        Return value must be a list of results. Each result is a
        :class:`~plinth.diagnostic_check.DiagnosticCheck` with a
        unique check_id, a user visible description of the test, the result,
        test parameters, and the component ID. The test result is a string
        enumeration from 'failed', 'passed', 'error', 'warning' and 'not_done'.

        Also see :meth:`.has_diagnostics`.
        """
        return []

    def has_diagnostics(self):
        """Return whether at least one diagnostic test is implemented.

        If this method return True, the :meth:`App.has_diagnostics`. also
        returns True.

        If a subclass of Component overrides the :meth:`.diagnose` method, it
        is assumed that it is for implementing diagnostic tests and this method
        returns True for such a component. Override this method if this default
        behavior does not fit the needs.
        """
        return self.__class__.diagnose is not Component.diagnose

    def repair(self, failed_checks: list) -> bool:
        """Try to fix failed diagnostics.

        The default implementation only requests re-run setup for the app.

        Returns whether the app setup should be re-run by the caller.

        This method should be overridden by components that implement
        diagnose(), if there is a known way to fix failed checks. The return
        value can be changed to False to avoid causing a re-run setup.

        failed_checks is a list of DiagnosticChecks related to this component
        that had failed or warning result.
        """
        return True


class FollowerComponent(Component):
    """Interface for an app component that follows other components.

    These components of the app don't determine if the app is enabled or not.

    """

    is_leader = False

    def __init__(self, component_id, is_enabled=False):
        """Initialize the component."""
        super().__init__(component_id)
        self._is_enabled = is_enabled

    def is_enabled(self):
        """Return whether the component is enabled."""
        return self._is_enabled

    def set_enabled(self, enabled):
        """Update the internal enabled state of the component."""
        self._is_enabled = enabled

    def enable(self):
        """Run operations to enable the component."""
        self._is_enabled = True

    def disable(self):
        """Run operations to disable the component."""
        self._is_enabled = False


class LeaderComponent(Component):
    """Interface for an app component that decides the state of the app.

    These components determine if the app is enabled or not.

    """

    is_leader = True

    def is_enabled(self):
        """Return if the component is enabled."""
        raise NotImplementedError


class Info(FollowerComponent):
    """Component to capture basic information about an app."""

    def __init__(self, app_id: str, version: int, is_essential: bool = False,
                 depends: list[str] | None = None, name: str | None = None,
                 icon: str | None = None, icon_filename: str | None = None,
                 description: list[str] | None = None,
                 manual_page: str | None = None,
                 clients: list[dict] | None = None,
                 donation_url: str | None = None,
                 tags: list[str] | None = None):
        """Store the basic properties of an app as a component.

        Each app must contain at least one component of this type to provide
        basic information about the app such as it's version number.

        Instead of polluting the list of properties of an app, this component
        stores them separately. This component can also be safely passed around
        to template etc. without exposing the methods of an app and without
        creating unnecessarily cyclic dependencies.

        'app_id' must be the unique ID of the app to which this information
        belongs.

        'version' is the monotonically increasing positive integer starting at
        1. It represents the version number of the app. It is used by the setup
        mechanism. When an app's version number is increased, the setup
        mechanism assumes that the setup() method of the app needs to run
        again. This is used to upgrade/change configuration/setup of a app
        when a new version of the app is deployed on users' machine.

        'is_essential' is a boolean that marks the app as mandatory for the
        basic functions of the system. If True, this app will be installed and
        setup during the first run of FreedomBox even before first setup wizard
        is shown to the user.

        'depends' is the list of other apps that this app depends on. Apps from
        this list are guaranteed to be initialized before initializing the app
        to which this component belongs.

        'name' is the user visible name of this app. It is shown as the title
        of the app in the list of apps and when viewing app details. It should
        be a lazily translated Django string.

        'icon' is the name of icon to use with this app from a predetermined
        list of icons. This is currently an icon class name from the Fork
        Awesome font. It is used when showing the app in the System section.
        Each app typically has either an 'icon' or 'icon_filename' property
        set.

        'icon_filename' is the name of the icon file, without the suffix, to be
        used with this app. A .svg file (used in the web interface) and a .png
        file (currently used by Android App) must be provided by the app. It is
        used in the primary app page and on the app listing page. Each app
        typically has either an 'icon' or 'icon_filename' property set.

        'description' is the user visible full description of the app. It is
        shown along in the app page along with other app details. It should be
        a list of lazily translated Django strings. Each string is rendered as
        a paragraph on the page. It may contain HTML <a> tags to provide links
        to external content.

        'manual_page' is the optional name of the page for this app in the user
        manual. If provided, a 'Learn more...' link appears in the app page for
        this app.

        'clients' is the list of applications that can be used with the
        services provided by this app. This is used to suggest installation of
        compatible clients on desktop, web and mobile. This is a list of
        dictionaries who structure is documented in plinth.clients.

        'donation_url' is a link to a webpage that describes how to
        donate to the upstream project.

        'tags' is a list of tags that describe the app. Tags help users to find
        similar apps or alternatives and discover use cases.

        """
        self.component_id = app_id + '-info'
        self.app_id = app_id
        self.version = version
        self.is_essential = is_essential
        self.depends = depends or []
        self.name = name
        self.icon = icon
        self.icon_filename = icon_filename
        self.description = description
        self.manual_page = manual_page
        self.clients = clients
        self.donation_url = donation_url
        self._tags = tags or []
        if clients:
            clients_module.validate(clients)

    @property
    def tags(self) -> list[str]:
        """Return a list of untranslated tags on the app.

        These can only be retrieved after Django has been configured.
        """
        # Store untranslated original strings instead of proxy objects
        from django.core.exceptions import ImproperlyConfigured
        from django.utils.functional import Promise
        from django.utils.translation import override
        try:
            with override(language=None):
                return [str(tag) for tag in self._tags]
        except ImproperlyConfigured:
            # Hack to allow apps to be instantiated without Django
            # initialization as required by privileged process.
            def _make_str(tag):
                """Return the string without casting."""
                if not isinstance(tag, Promise):
                    return tag

                # Django 4.2
                if hasattr(tag, '_proxy____args'):
                    return tag._proxy____args[0]

                # Django 5.x
                return tag._args[0]

            return [_make_str(tag) for tag in self._tags]


class EnableState(LeaderComponent):
    """A component to hold the enable state of an app using a simple flag.

    The flag is stored in the FreedomBox service database. This component
    should only be used if an app does not have any other way to determine if
    it is enabled or disabled by examining the state of the system. Typical
    apps have daemons, web server configuration, etc. and the enabled/disabled
    state of those determine the enabled/disabled state of the entire app. If
    an does not have any such system state, then this component may be used to
    provide enable/disable functionality for the app.
    """

    @property
    def key(self):
        """Return the storage key."""
        if not self.app_id:
            raise KeyError('Component must be added to an app before use.')

        return f'{self.app_id}_enable'

    def is_enabled(self):
        """Return whether the app/component is enabled."""
        from plinth import kvstore
        return kvstore.get_default(self.key, False)

    def enable(self):
        """Store that the app/component is enabled."""
        from plinth import kvstore
        kvstore.set(self.key, True)

    def disable(self):
        """Store that the app/component is disabled."""
        from plinth import kvstore
        kvstore.set(self.key, False)


def apps_init():
    """Create apps by constructing them with components."""
    if App.list():
        return  # Apps have already been initialized

    from . import module_loader  # noqa  # Avoid circular import
    for module_name, module in module_loader.loaded_modules.items():
        _initialize_module(module_name, module)

    _sort_apps()

    logger.info('Initialized apps - %s', ', '.join(
        (app.app_id for app in App.list())))


def _sort_apps():
    """Sort apps list according to their essential/dependency order."""
    apps = App._all_apps
    ordered_modules = []
    remaining_apps = dict(apps)  # Make a copy
    # Place all essential modules ahead of others in module load order
    sorted_apps = sorted(
        apps, key=lambda app_id: not App.get(app_id).info.is_essential)
    for app_id in sorted_apps:
        if app_id not in remaining_apps:
            continue

        app = remaining_apps.pop(app_id)
        try:
            _insert_apps(app_id, app, remaining_apps, ordered_modules)
        except KeyError:
            logger.error('Unsatified dependency for app - %s', app_id)

    new_all_apps = collections.OrderedDict()
    for app_id in ordered_modules:
        new_all_apps[app_id] = apps[app_id]

    App._all_apps = new_all_apps


def _insert_apps(app_id, app, remaining_apps, ordered_apps):
    """Insert apps into a list based on dependency order."""
    if app_id in ordered_apps:
        return

    for dependency in app.info.depends:
        if dependency in ordered_apps:
            continue

        try:
            app = remaining_apps.pop(dependency)
        except KeyError:
            logger.error('Not found or circular dependency - %s, %s', app_id,
                         dependency)
            raise

        _insert_apps(dependency, app, remaining_apps, ordered_apps)

    ordered_apps.append(app_id)


def _initialize_module(module_name, module):
    """Perform initialization on all apps in a module."""
    try:
        module_classes = inspect.getmembers(module, inspect.isclass)
        app_classes = [
            cls for _, cls in module_classes if issubclass(cls, App)
        ]
        for app_class in app_classes:
            app_class()
    except Exception as exception:
        logger.exception('Exception while running init for module %s: %s',
                         module_name, exception)
        if cfg.develop:
            raise


def apps_post_init():
    """Run post initialization on each app."""
    for app in App.list():
        try:
            app.post_init()
            if not app.needs_setup() and app.is_enabled():
                app.set_enabled(True)
        except Exception as exception:
            logger.exception('Exception while running post init for %s: %s',
                             app.app_id, exception)
            if cfg.develop:
                raise

    logger.debug('App initialization completed.')
    post_app_loading.send_robust(sender="app")



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Views for radicale module.
"""

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth.views import AppView

from . import privileged
from .forms import RadicaleForm


class RadicaleAppView(AppView):
    """A specialized view for configuring radicale service."""
    form_class = RadicaleForm
    app_id = 'radicale'

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial['access_rights'] = privileged.get_rights_value()
        return initial

    def form_valid(self, form):
        """Change the access control of Radicale service."""
        data = form.cleaned_data
        if privileged.get_rights_value() != data['access_rights']:
            privileged.configure(data['access_rights'])
            messages.success(self.request,
                             _('Access rights configuration updated'))
        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for configuring Ejabberd.
"""

from django import forms
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.modules import ejabberd
from plinth.modules.coturn.forms import turn_uris_validator
from plinth.modules.names.components import DomainName
from plinth.utils import format_lazy


class EjabberdForm(forms.Form):
    """Ejabberd configuration form."""
    domain_names = forms.MultipleChoiceField(
        label=_('Domain names'), widget=forms.CheckboxSelectMultiple,
        help_text=_(
            'Domains to be used by ejabberd. Note that user accounts are '
            'unique for each domain, and migrating users to a new domain name '
            'is not yet implemented.'), choices=[])

    MAM_enabled = forms.BooleanField(
        label=_('Enable Message Archive Management'), required=False,
        help_text=format_lazy(
            _('If enabled, your {box_name} will store chat message histories. '
              'This allows synchronization of conversations between multiple '
              'clients, and reading the history of a multi-user chat room. '
              'It depends on the client settings whether the histories are '
              'stored as plain text or encrypted.'), box_name=_(cfg.box_name)))

    enable_managed_turn = forms.BooleanField(
        label=_('Automatically manage audio/video call setup'), required=False,
        help_text=format_lazy(
            _('Configures the local <a href={coturn_url}>coturn</a> app as '
              'the STUN/TURN server for ejabberd. Disable this if you '
              'want to use a different STUN/TURN server.'),
            coturn_url=reverse_lazy('coturn:index')))

    # STUN/TURN server setup
    turn_uris = forms.CharField(
        label=_('STUN/TURN Server URIs'), required=False, strip=True,
        widget=forms.Textarea(attrs={'rows': 4}),
        help_text=_('List of public URIs of the STUN/TURN server, one on each '
                    'line.'), validators=[turn_uris_validator])

    shared_secret = forms.CharField(
        label=_('Shared Authentication Secret'), required=False, strip=True,
        help_text=_('Shared secret used to compute passwords for the '
                    'TURN server.'))

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # Start with any existing domains from ejabberd configuration.
        domains = ejabberd.get_domains()

        # Add other domains that can be configured.
        for domain in DomainName.list_names():
            if domain not in domains:
                domains.append(domain)

        self.fields['domain_names'].choices = zip(domains, domains)

    def clean_turn_uris(self):
        """Normalize newlines in URIs."""
        data = self.cleaned_data['turn_uris']
        return '\n'.join([uri.strip() for uri in data.splitlines()])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Framework to run specified actions with elevated privileges."""

import functools
import importlib
import inspect
import io
import json
import logging
import pathlib
import socket
import threading
import traceback
import types
import typing

from plinth import cfg, module_loader

logger = logging.getLogger(__name__)

socket_path = '/run/freedombox/privileged.socket'

thread_storage = threading.local()


# An alias for 'str' to mark some strings as sensitive. Sensitive strings are
# not logged. Use 'type secret_str = str' when Python 3.11 support is no longer
# needed.
class secret_str(str):
    pass


def privileged(func):
    """Mark a method as allowed to be run as privileged method.

    This decorator is to mark any method as needing to be executed with
    superuser privileges. This is necessary because the primary FreedomBox
    service daemon runs as a regular user and has no special privileges. When
    performing system operations, FreedomBox service will either communicate
    with privileged daemons such as NetworkManager and systemd, or spawns a
    separate process with higher privileges. When spawning a separate process
    all the action parameters need to serialized, communicated to the process
    and then de-serialized inside the process. The return value also need to
    undergo such serialization and de-serialization. This decorator makes this
    task simpler.

    A call to a decorated method will be serialized into a sudo call (or later
    into a D-Bus call). The method arguments are turned to JSON and method is
    called with superuser privileges. As arguments are de-serialized, they are
    verified for type before the actual call as superuser. Return values are
    serialized and returned where they are de-serialized. Exceptions are also
    serialized and de-serialized. The decorator wrapper code will either return
    the value or raise exception.

    For a method to be decorated, the method must have type annotations for all
    of its parameters and should not use keyword-only arguments. It must also
    be in a module named privileged.py directly under the application similar
    to models.py, views.py and urls.py. Currently supported types are bool,
    int, float, str, dict/Dict, list/List, Optional and Union.

    Privileged methods many not output to the stdout as it interferes
    with the serialization and de-serialization process.
    """
    setattr(func, '_privileged', True)

    _check_privileged_action_arguments(func)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if getattr(func, '_skip_privileged_call', False):
            return func(*args, **kwargs)

        module_name = _get_privileged_action_module_name(func)
        action_name = func.__name__
        return run_privileged_method(func, module_name, action_name, args,
                                     kwargs)

    return wrapper


def _read_from_server(client_socket: socket.socket) -> bytes:
    """Read everything from a socket and return the data."""
    response = b''
    while True:
        chunk = client_socket.recv(4096)
        if not chunk:
            break

        response += chunk

    if not response:
        raise ConnectionError('Server returned empty response')

    return json.loads(response)


def _request_to_server(request: dict) -> socket.socket:
    """Connect to the server and make a request."""
    request_string = json.dumps(request)
    client_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        client_socket.connect(socket_path)
        client_socket.sendall(request_string.encode('utf-8'))
        # Close the write end of the socket signaling an EOF and no more data
        # will be sent.
        client_socket.shutdown(socket.SHUT_WR)
    except Exception:
        client_socket.close()
        raise

    return client_socket


def run_privileged_method(func, module_name, action_name, args, kwargs):
    """Execute a privileged method using a server."""
    run_in_background = kwargs.pop('_run_in_background', False)
    raw_output = kwargs.pop('_raw_output', False)
    log_error = kwargs.pop('_log_error', True)

    if func:
        _log_action(func, module_name, action_name, args, kwargs,
                    run_in_background)

    request = {
        'module': module_name,
        'action': action_name,
        'args': args,
        'kwargs': kwargs,
    }
    if raw_output:
        request['raw_output'] = raw_output

    if not log_error:
        request['log_error'] = False

    client_socket = _request_to_server(request)

    if raw_output:

        def _reader_func():
            while True:
                chunk = client_socket.recv(4096)
                if chunk:
                    yield chunk
                else:
                    break

            client_socket.close()

        return _reader_func()

    args = (func, module_name, action_name, args, kwargs, log_error,
            client_socket)
    if not run_in_background:
        return _wait_for_server_response(*args)

    read_thread = threading.Thread(target=_wait_for_server_response, args=args)
    read_thread.start()


def _wait_for_server_response(func, module_name, action_name, args, kwargs,
                              log_error, client_socket):
    """Wait for the server to respond and process the response."""
    try:
        return_value = _read_from_server(client_socket)
    except json.JSONDecodeError:
        logger.error('Error decoding action return value %s..%s(*%s, **%s)',
                     module_name, action_name, args, kwargs)
        raise
    finally:
        client_socket.close()

    if return_value['result'] == 'success':
        return return_value['return']

    module = importlib.import_module(return_value['exception']['module'])
    exception_class = getattr(module, return_value['exception']['name'])
    exception = exception_class(*return_value['exception']['args'])
    exception.stdout = return_value['exception'].get('stdout', b'').encode()
    exception.stderr = return_value['exception'].get('stderr', b'').encode()

    def _get_html_message():
        """Return an HTML format error that can be shown in messages."""
        from django.utils.html import format_html

        formatted_args = _format_args(func, args, kwargs)
        exception_args, stdout, stderr, traceback = _format_error(
            exception, return_value)
        return format_html('Error running action: {}..{}({}): {}({})\n{}{}{}',
                           module_name, action_name, formatted_args,
                           return_value['exception']['name'], exception_args,
                           stdout, stderr, traceback)

    if func:
        exception.get_html_message = _get_html_message

    if log_error and func:
        formatted_args = _format_args(func, args, kwargs)
        exception_args, stdout, stderr, traceback = _format_error(
            exception, return_value)
        logger.error('Error running action %s..%s(%s): %s(%s)\n'
                     '%s%s%s', module_name, action_name, formatted_args,
                     return_value['exception']['name'], exception_args, stdout,
                     stderr, traceback)

    raise exception


class ProcessBufferedReader(io.BufferedReader):
    """Improve performance of buffered binary streaming.

    Read from the stdout of a process and close the process after reading is
    completed.

    Iteration calls __next__ over the BufferedReader holding process.stdout.
    However, this seems to call readline() which looks for \n in binary data
    which leads to short unpredictably sized chunks which in turn lead to
    severe performance degradation. So, overwrite this and call read() which is
    better geared for handling binary data.
    """

    def __init__(self, process, extra_cleanup_func=None, *args, **kwargs):
        """Store the process object."""
        super().__init__(process.stdout, *args, **kwargs)
        self.process = process
        self.extra_cleanup_func = extra_cleanup_func

    def __next__(self):
        """Override to call read() instead of readline()."""
        chunk = self.read(io.DEFAULT_BUFFER_SIZE)
        if not chunk:
            self._cleanup_func()

            raise StopIteration

        return chunk

    def _cleanup_func(self):
        """After the process has been read from, cleanup the process."""
        try:
            if self.process.stdout:
                self.process.stdout.close()

            if self.process.stderr:
                self.process.stderr.close()

            self.process.wait(30)
            if self.extra_cleanup_func:
                self.extra_cleanup_func()
        except Exception:
            logger.exception('Closing process failed after raw output')


def _format_args(func, args, kwargs):
    """Return a loggable representation of arguments."""
    argspec = inspect.getfullargspec(func)
    if len(args) > len(argspec.args):
        raise SyntaxError('Too many arguments')

    args_str_list = []
    for arg_index, arg_value in enumerate(args):
        arg_name = argspec.args[arg_index]
        if argspec.annotations[arg_name] in [secret_str, secret_str | None]:
            value = '****'
        else:
            value = json.dumps(arg_value)

        args_str_list.append(value)

    kwargs_str_list = []
    for arg_name, arg_value in kwargs.items():
        if argspec.annotations[arg_name] in [secret_str, secret_str | None]:
            value = "****"
        else:
            value = json.dumps(arg_value)

        kwargs_str_list.append(f'{arg_name}=' + value)

    return ', '.join(args_str_list + kwargs_str_list)


def _format_error(exception, return_value):
    """Log the exception in a readable manner."""
    exception_args = ', '.join([json.dumps(arg) for arg in exception.args])

    stdout = exception.stdout.decode()
    if stdout:
        lines = stdout.split('\n')
        lines = lines[:-1] if not lines[-1] else lines
        stdout = '\n'.join(('│ ' + line for line in lines))
        stdout = 'Stdout:\n' + stdout + '\n'

    stderr = exception.stderr.decode()
    if stderr:
        lines = stderr.split('\n')
        lines = lines[:-1] if not lines[-1] else lines
        stderr = '\n'.join(('║ ' + line for line in lines))
        stderr = 'Stderr:\n' + stderr + '\n'

    traceback = return_value['exception']['traceback']
    if traceback:
        all_lines = []
        for entry in traceback:
            lines = entry.split('\n')
            all_lines += lines[:-1] if not lines[-1] else lines

        traceback = '\n'.join(('╞ ' + line for line in all_lines))
        traceback = 'Action traceback:\n' + traceback + '\n'

    return (exception_args, stdout, stderr, traceback)


def _check_privileged_action_arguments(func):
    """Check that a privileged action has well defined types."""
    argspec = inspect.getfullargspec(func)
    if (argspec.varargs or argspec.varkw or argspec.kwonlyargs
            or argspec.kwonlydefaults):
        raise SyntaxError('Actions must not have variable args')

    for arg in argspec.args:
        if arg not in argspec.annotations:
            raise SyntaxError('All arguments must be annotated')

    for arg_name, arg_value in argspec.annotations.items():
        for keyword in ('password', 'passphrase', 'secret'):
            if keyword in arg_name:
                if arg_value not in [secret_str, secret_str | None]:
                    raise SyntaxError(
                        f'Argument {arg_name} should likely be a "secret_str"')


def _get_privileged_action_module_name(func):
    """Figure out the module name of a privileged action."""
    module_name = func.__module__
    while module_name:
        module_name, _, last = module_name.rpartition('.')
        if last == 'privileged':
            break

    if not module_name:
        raise ValueError('Privileged actions must be placed under a '
                         'package/module named privileged')

    return module_name.rpartition('.')[2]


def _log_action(func, module_name, action_name, args, kwargs,
                run_in_background):
    """Log an action in a compact format."""
    prompt = '»'
    suffix = '&' if run_in_background else ''
    formatted_args = _format_args(func, args, kwargs)
    logger.info('%s %s..%s(%s) %s', prompt, module_name, action_name,
                formatted_args, suffix)


class JSONEncoder(json.JSONEncoder):
    """Handle to special types that default JSON encoder does not."""

    def default(self, obj):
        """Handle special object types."""
        # When subprocess.call() fails and one of the arguments is a Path-like
        # object, the exception also contains a Path-like object.
        if isinstance(obj, pathlib.Path):
            return str(obj)

        return super().default(obj)


def _setup_thread_storage():
    """Setup collection of stdout/stderr from any process in this thread."""
    thread_storage.stdout = b''
    thread_storage.stderr = b''


def _clear_thread_storage():
    """Cleanup memory used for stdout/stderr from processes in this thread.

    Python documentation is silent on whether thread local storage will be
    cleaned up after a thread terminates.
    """
    thread_storage.stdout = None
    thread_storage.stderr = None


def get_return_value_from_exception(exception):
    """Return the value to return from server when an exception is raised."""
    return_value = {
        'result': 'exception',
        'exception': {
            'module': type(exception).__module__,
            'name': type(exception).__name__,
            'args': exception.args,
            'traceback': traceback.format_tb(exception.__traceback__),
            'stdout': getattr(thread_storage, 'stdout', b'').decode(),
            'stderr': getattr(thread_storage, 'stderr', b'').decode(),
        }
    }
    return return_value


def privileged_handle_json_request(
        request_string: str) -> str | io.BufferedReader:
    """Parse arguments for the program spawned as a privileged action."""

    def _parse_request() -> dict:
        """Return a JSON parsed and validated request."""
        try:
            request = json.loads(request_string)
        except json.JSONDecodeError:
            raise SyntaxError('Invalid JSON in request')

        required_parameters = [('module', str), ('action', str),
                               ('args', list), ('kwargs', dict)]

        for parameter, expected_type in required_parameters:
            if parameter not in request:
                raise TypeError(f'Missing required parameter "{parameter}"')
            if not isinstance(request[parameter], expected_type):
                raise TypeError(f'Parameter "{parameter}" must be of type'
                                f'{expected_type.__name__}')

        if 'raw_output' in request and not isinstance(request['raw_output'],
                                                      bool):
            raise TypeError('Incorrect "raw_output" parameter')

        if 'log_error' in request and not isinstance(request['log_error'],
                                                     bool):
            raise TypeError('Incorrect "log_error" parameter')

        return request

    try:
        request = _parse_request()
        log_error = request.get('log_error', True)
        arguments = {'args': request['args'], 'kwargs': request['kwargs']}
        _setup_thread_storage()
        return_value = _privileged_call(request['module'], request['action'],
                                        arguments, log_error)

        if isinstance(return_value, io.BufferedReader):
            raw_output = request.get('raw_output', False)
            if not raw_output:
                raise TypeError('Invalid call to raw output API.')

            return return_value
    except (PermissionError, SyntaxError, TypeError, Exception) as exception:
        if isinstance(exception, (PermissionError, SyntaxError, TypeError)):
            logger.error(exception.args[0])
        else:
            logger.exception(exception)

        return_value = get_return_value_from_exception(exception)

    _clear_thread_storage()

    return json.dumps(return_value, cls=JSONEncoder)


def _privileged_call(module_name, action_name, arguments, log_error=True):
    """Import the module and run action as superuser"""
    if '.' in module_name:
        raise SyntaxError('Invalid module name')

    cfg.read()
    if module_name == 'plinth':
        import_path = 'plinth'
    else:
        try:
            import_path = module_loader.get_module_import_path(module_name)
        except FileNotFoundError as exception:
            raise SyntaxError('Specified module not found') from exception

    try:
        module = importlib.import_module(import_path + '.privileged')
    except ModuleNotFoundError as exception:
        raise SyntaxError('Specified module not found') from exception

    try:
        action = getattr(module, action_name)
    except AttributeError as exception:
        raise SyntaxError('Specified action not found') from exception

    if not getattr(action, '_privileged', None):
        raise SyntaxError('Specified action is not privileged action')

    # Get the original function that may have been wrapped/decorated multiple
    # times
    func = action
    while True:
        try:
            func = getattr(func, '__wrapped__')
        except AttributeError:
            break

    _privileged_assert_valid_arguments(func, arguments)

    _log_action(func, module_name, action_name, arguments['args'],
                arguments['kwargs'], run_in_background=False)
    try:
        return_values = func(*arguments['args'], **arguments['kwargs'])
        if isinstance(return_values, io.BufferedReader):
            return return_values

        return_value = {'result': 'success', 'return': return_values}
    except Exception as exception:
        return_value = get_return_value_from_exception(exception)
        if log_error:
            logger.exception(
                'Error running action: %s..%s(..): %s\nstdout:\n%s\n'
                'stderr:\n%s\n', module_name, action_name, exception,
                return_value['exception']['stdout'],
                return_value['exception']['stderr'])

    return return_value


def _privileged_assert_valid_arguments(func, arguments):
    """Check the names, types and completeness of the arguments passed."""
    # Check if arguments match types
    if not isinstance(arguments, dict):
        raise SyntaxError('Invalid arguments format')

    if 'args' not in arguments or 'kwargs' not in arguments:
        raise SyntaxError('Invalid arguments format')

    args = arguments['args']
    kwargs = arguments['kwargs']
    if not isinstance(args, list) or not isinstance(kwargs, dict):
        raise SyntaxError('Invalid arguments format')

    argspec = inspect.getfullargspec(func)
    if len(args) + len(kwargs) > len(argspec.args):
        raise SyntaxError('Too many arguments')

    no_defaults = len(argspec.args)
    if argspec.defaults:
        no_defaults -= len(argspec.defaults)

    for key in argspec.args[len(args):no_defaults]:
        if key not in kwargs:
            raise SyntaxError(f'Argument not provided: {key}')

    for key, value in kwargs.items():
        if key not in argspec.args:
            raise SyntaxError(f'Unknown argument: {key}')

        if argspec.args.index(key) < len(args):
            raise SyntaxError(f'Duplicate argument: {key}')

        _privileged_assert_valid_type(f'arg {key}', value,
                                      argspec.annotations[key])

    for index, arg in enumerate(args):
        annotation = argspec.annotations[argspec.args[index]]
        _privileged_assert_valid_type(f'arg #{index}', arg, annotation)


def _privileged_assert_valid_type(arg_name, value, annotation):
    """Assert that the type of argument value matches the annotation."""
    if annotation == typing.Any:
        return

    NoneType = type(None)
    if annotation == NoneType:
        if value is not None:
            raise TypeError('Expected None for {arg_name}')

        return

    basic_types = {bool, int, str, float}
    if annotation in basic_types:
        if not isinstance(value, annotation):
            raise TypeError(
                f'Expected type {annotation.__name__} for {arg_name}')

        return

    # secret_str should be a regular string
    if annotation == secret_str:
        if not isinstance(value, str):
            raise TypeError(f'Expected type str for {arg_name}')

        return

    # 'int | str' or 'typing.Union[int, str]'
    if (isinstance(annotation, types.UnionType)
            or getattr(annotation, '__origin__', None) == typing.Union):
        for arg in annotation.__args__:
            try:
                _privileged_assert_valid_type(arg_name, value, arg)
                return
            except TypeError:
                pass

        raise TypeError(f'Expected one of unioned types for {arg_name}')

    # 'list[int]' or 'typing.List[int]'
    if getattr(annotation, '__origin__', None) == list:
        if not isinstance(value, list):
            raise TypeError(f'Expected type list for {arg_name}')

        for index, inner_item in enumerate(value):
            _privileged_assert_valid_type(f'{arg_name}[{index}]', inner_item,
                                          annotation.__args__[0])

        return

    # 'list[dict]' or 'typing.List[dict]'
    if getattr(annotation, '__origin__', None) == dict:
        if not isinstance(value, dict):
            raise TypeError(f'Expected type dict for {arg_name}')

        for inner_key, inner_value in value.items():
            _privileged_assert_valid_type(f'{arg_name}[{inner_key}]',
                                          inner_key, annotation.__args__[0])
            _privileged_assert_valid_type(f'{arg_name}[{inner_value}]',
                                          inner_value, annotation.__args__[1])

        return

    raise TypeError('Unsupported annotation type')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django views for Feather Wiki."""

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView

from plinth import app as app_module
from plinth import views
from plinth.modules import featherwiki
from plinth.views import messages_error

from . import privileged
from .forms import CreateWikiForm, RenameWikiForm, UploadWikiForm

DUPLICATE_FILE_ERROR = _('A wiki file with the given name already exists.')


class FeatherWikiAppView(views.AppView):
    """Serve configuration page."""

    app_id = 'featherwiki'
    template_name = 'featherwiki_configure.html'

    def get_context_data(self, *args, **kwargs):
        """Add wikis to the context data."""
        context = super().get_context_data(*args, **kwargs)
        context['wikis'] = featherwiki.get_wiki_list()
        return context


class CreateWikiView(SuccessMessageMixin, FormView):
    """View to create a new repository."""

    form_class = CreateWikiForm
    prefix = 'featherwiki'
    template_name = 'form.html'
    success_url = reverse_lazy('featherwiki:index')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Create Wiki')
        return context

    def form_valid(self, form):
        """Create the repository on valid form submission."""
        try:
            privileged.create_wiki(form.cleaned_data['name'])
            self.success_message = _('Wiki created.')
        except ValueError:
            messages.error(self.request, DUPLICATE_FILE_ERROR)
        except Exception as error:
            messages_error(self.request,
                           _('An error occurred while creating the wiki.'),
                           error)

        return super().form_valid(form)


class RenameWikiView(SuccessMessageMixin, FormView):
    """View to edit an existing repository."""

    form_class = RenameWikiForm
    prefix = 'featherwiki'
    template_name = 'form.html'
    success_url = reverse_lazy('featherwiki:index')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Rename Wiki')
        return context

    def form_valid(self, form):
        """Rename the wiki on valid form submission."""
        try:
            privileged.rename_wiki(self.kwargs['old_name'],
                                   form.cleaned_data['new_name'])
            self.success_message = _('Wiki renamed.')
        except ValueError:
            messages.error(self.request, DUPLICATE_FILE_ERROR)
        except Exception as error:
            messages_error(self.request,
                           _('An error occurred while renaming the wiki.'),
                           error)

        return super().form_valid(form)


class UploadWikiView(SuccessMessageMixin, FormView):
    """View to upload an existing wiki file."""

    form_class = UploadWikiForm
    prefix = 'featherwiki'
    template_name = 'featherwiki_upload_file.html'
    success_url = reverse_lazy('featherwiki:index')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Upload Wiki File')
        return context

    def form_valid(self, form):
        """Add the wiki file on valid form submission."""
        try:
            uploaded_file = self.request.FILES['featherwiki-file']
            privileged.add_wiki_file(uploaded_file.name,
                                     uploaded_file.temporary_file_path())
            self.success_message = _('Wiki file added.')
        except FileExistsError:
            messages.error(self.request, DUPLICATE_FILE_ERROR)
        except Exception as error:
            messages_error(self.request, _('Failed to add wiki file.'), error)
            return redirect(reverse_lazy('featherwiki:index'))

        return super().form_valid(form)


def delete(request, name):
    """Handle deleting wikis, showing a confirmation dialog first.

    On GET, display a confirmation page.
    On POST, delete the wiki.
    """
    app = app_module.App.get('featherwiki')
    if request.method == 'POST':
        try:
            privileged.delete_wiki(name)
            messages.success(request, _('{name} deleted.').format(name=name))
        except Exception as error:
            messages_error(request,
                           _('Could not delete {name}').format(name=name),
                           error)

        return redirect(reverse_lazy('featherwiki:index'))

    return TemplateResponse(request, 'featherwiki_delete.html', {
        'title': app.info.name,
        'name': name
    })



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for privoxy.
"""

from django.utils.translation import gettext_lazy as _

backup: dict = {}

tags = [_('Ad blocker'), _('Proxy server'), _('Local network')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the OpenID Connect module.

All the '/freedombox/o' URLs are implemented in this module by including them
from django-oauth-toolkit. However, they are included in plinth/urls.py instead
of here because FreedomBox module loading logic automatically namespaces the
URL names. This causes problems when metadata view tries to resolve URLs.

/.well-known/openid-configuration is proxied to
/freedombox/o/.well-known/openid-configuration by Apache2. Similarly,
/.well-known/jwks.json is proxied to /freedombox/o/.well-known/jwks.json.

Important URLs:

- /freedombox/o is the primary URL for identity provider.

- /freedombox/o/.well-known/openid-configuration is the way to discover
additional URLs (such as ./authorize and ./token) needed for OIDC to work.

- /freedombox/o/authorize is used to start the authorization process and get an
authorization code grant.

- /freedombox/o/token is used to get access token and refresh token using the
authorization code. It is also used to get a new access token using the refresh
token.

- /freedombox/o/userinfo provides the claims such as 'sub', 'email',
'freedombox_groups' using an access token.
"""

urlpatterns: list = []



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app to configure minidlna.
"""
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.daemon import Daemon
from plinth.modules import firewall
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages, install
from plinth.utils import Version

from . import manifest, privileged

_description = [
    _('MiniDLNA is a simple media server software, with the aim of being '
      'fully compliant with DLNA/UPnP-AV clients. '
      'The MiniDLNA daemon serves media files '
      '(music, pictures, and video) to clients on a network. '
      'DLNA/UPnP is zero configuration protocol and is compliant '
      'with any device passing the DLNA Certification like portable '
      'media players, Smartphones, Televisions, and gaming systems ('
      'such as PS3 and Xbox 360) or applications such as totem and Kodi.')
]

SYSTEM_USER = 'minidlna'


class MiniDLNAApp(app_module.App):
    """Freedombox app managing miniDlna."""

    app_id = 'minidlna'

    _version = 6

    def __init__(self) -> None:
        """Initialize the app components."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('MiniDLNA'), icon_filename='minidlna',
                               description=_description,
                               manual_page='MiniDLNA',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-minidlna', name=info.name,
                              icon=info.icon_filename, tags=info.tags,
                              url_name='minidlna:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-minidlna', info.name, description=info.description,
            icon=info.icon_filename,
            configure_url=reverse_lazy('minidlna:index'), tags=info.tags,
            login_required=True)
        self.add(shortcut)

        packages = Packages('packages-minidlna', ['minidlna'])
        self.add(packages)

        firewall_minidlna = Firewall('firewall-minidlna', info.name,
                                     ports=['minidlna',
                                            'ssdp'], is_external=False)
        self.add(firewall_minidlna)

        daemon = Daemon('daemon-minidlna', 'minidlna')
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-minidlna',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if old_version == 3:
            # Version 3 of the app incorrectly declared port 8200 for firewall
            # local protection.
            firewall.remove_passthrough('ipv6', '-A', 'INPUT', '-p', 'tcp',
                                        '--dport', '8200', '-j', 'REJECT')
            firewall.remove_passthrough('ipv4', '-A', 'INPUT', '-p', 'tcp',
                                        '--dport', '8200', '-j', 'REJECT')

        if old_version and old_version <= 5:
            # Remove minidlna LDAP group and disable minidlna apache config
            from plinth.modules.apache import privileged as apache_privileged
            from plinth.modules.users import privileged as users_privileged

            users_privileged.remove_group('minidlna')
            apache_privileged.disable('minidlna-freedombox', 'config')

            # Restart app to reload firewall
            if self.is_enabled():
                self.disable()
                self.enable()

        if not old_version:
            self.enable()

    def force_upgrade(self, packages):
        """Force upgrade minidlna to resolve conffile prompt."""
        if 'minidlna' not in packages:
            return False

        # Allow upgrade from 1.3.0 (bookworm) to 1.3.3 (trixie) and beyond
        # 1.3.x.
        package = packages['minidlna']
        if Version(package['new_version']) > Version('1.4~'):
            return False

        media_dir = privileged.get_media_dir()
        install(['minidlna'], force_configuration='new')
        privileged.set_media_dir(media_dir)

        return True



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Set and get postfix configuration using postconf.

See: http://www.postfix.org/postconf.1.html
See: http://www.postfix.org/master.5.html
See: http://www.postfix.org/postconf.5.html
"""

import re
import subprocess
from dataclasses import dataclass

from plinth import action_utils


@dataclass
class Service:  # NOQA, pylint: disable=too-many-instance-attributes
    """Representation of a postfix daemon and its options."""
    service: str
    type_: str
    private: str
    unpriv: str
    chroot: str
    wakeup: str
    maxproc: str
    command: str
    options: dict[str, str]

    def __str__(self) -> str:
        parts = [
            self.service, self.type_, self.private, self.unpriv, self.chroot,
            self.wakeup, self.maxproc, self.command
        ]
        for key, value in self.options.items():
            _validate_key(key)
            _validate_value(value)
            parts.extend(['-o', f'{key}={value}'])

        return ' '.join(parts)


def get_config(keys: list) -> dict:
    """Get postfix configuration using the postconf command."""
    for key in keys:
        _validate_key(key)

    args = ['/sbin/postconf']
    for key in keys:
        args.append(key)

    output = _run(args)
    result = {}
    lines: list[str] = list(filter(None, output.split('\n')))
    for line in lines:
        key, sep, value = line.partition('=')
        if not sep:
            raise ValueError('Invalid output detected')

        result[key.strip()] = value.strip()

    if set(keys) != set(result.keys()):
        raise ValueError('Some keys were missing from the output')

    return result


def set_config(config: dict, flag=None):
    """Set postfix configuration using the postconf command."""
    if not config:
        return

    for key, value in config.items():
        _validate_key(key)
        _validate_value(value)

    args = ['/sbin/postconf']
    if flag:
        args.append(flag)

    for key, value in config.items():
        args.append('{}={}'.format(key, value))

    _run(args)


def set_master_config(service: Service):
    """Set daemons and their options in postfix master.cf."""
    service_key = service.service + '/' + service.type_
    set_config({service_key: str(service)}, '-M')


def parse_maps(raw_value):
    """Parse postfix configuration values that are maps."""
    if '{' in raw_value or '}' in raw_value:
        raise ValueError('Unsupported map list format')

    value_list = []
    for segment in raw_value.split(','):
        for sub_segment in segment.strip().split(' '):
            sub_segment = sub_segment.strip()
            if sub_segment:
                value_list.append(sub_segment)

    return value_list


def _run(args):
    """Run process. Capture and return standard output as a string.

    Raise a RuntimeError on non-zero exit codes.
    """
    try:
        result = action_utils.run(args, check=True)
        return result.stdout.decode()
    except subprocess.SubprocessError as subprocess_error:
        raise RuntimeError('Subprocess failed') from subprocess_error
    except UnicodeDecodeError as unicode_error:
        raise RuntimeError('Unicode decoding failed') from unicode_error


def _validate_key(key):
    """Validate postconf key format or raise ValueError."""
    if not re.match(r'^[a-zA-Z][a-zA-Z0-9_/]*$', key):
        raise ValueError('Invalid postconf key format')


def _validate_value(value):
    """Validate postconf value format or raise ValueError."""
    for char in value:
        if ord(char) < 32:
            raise ValueError('Value contains control characters')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Miscellaneous utility methods.
"""

import gzip
import importlib
import os
import random
import re
import string

import markupsafe
import ruamel.yaml
from django.utils.functional import lazy

from plinth.version import Version  # noqa


def import_from_gi(library, version):
    """Import and return a GObject introspection library."""
    try:
        import gi as package
        package_name = 'gi'
    except ImportError:
        import pgi as package
        package_name = 'pgi'

    package.require_version(library, version)

    return importlib.import_module(package_name + '.repository.' + library)


def _format_lazy(string, *args, **kwargs):
    """Lazily format a lazy string."""
    allow_markup = kwargs.pop('allow_markup', False)
    string = str(string)
    string = string.format(*args, **kwargs)
    if allow_markup:
        string = markupsafe.Markup(string)

    return string


format_lazy = lazy(_format_lazy, str)


def non_admin_view(func):
    """Decorator to mark a view as accessible by non-admin users."""
    setattr(func, 'IS_NON_ADMIN', True)
    return func


def user_group_view(func, group_name):
    """Decorator to mark a view as accessible by admin or group users."""
    setattr(func, 'GROUP_NAME', group_name)
    return func


def is_valid_user_name(user_name):
    """Check if the given username is valid.

    Note: Debian is VERY flexible with user names.
    """
    if 32 < len(user_name):
        return False

    if user_name.startswith('-'):
        return False

    for forbidden in (' \n\t/\\:'):
        if forbidden in user_name:
            return False

    return True


def is_user_admin(request, cached=False):
    """Return whether user is an administrator."""
    if not request.user.is_authenticated:
        return False

    if 'cache_user_is_admin' in request.session and cached:
        return request.session['cache_user_is_admin']

    user_is_admin = request.user.groups.filter(name='admin').exists()
    request.session['cache_user_is_admin'] = user_is_admin
    return user_is_admin


class YAMLFile:
    """A context management class for updating YAML files"""

    def __init__(self, yaml_file):
        """Return a context object for the YAML file.

        Parameters:
        yaml_file - the YAML file to update.
        updating the YAML file.
        """
        self.yaml_file = yaml_file
        self.conf = None

        self.yaml = ruamel.yaml.YAML()
        self.yaml.preserve_quotes = True

    def __enter__(self):
        with open(self.yaml_file, 'r', encoding='utf-8') as intro_conf:
            if not self.is_file_empty():
                self.conf = self.yaml.load(intro_conf)
            else:
                self.conf = {}

            return self.conf

    def __exit__(self, type_, value, traceback):
        if not traceback:
            with open(self.yaml_file, 'w', encoding='utf-8') as intro_conf:
                self.yaml.dump(self.conf, intro_conf)

    def is_file_empty(self):
        return os.stat(self.yaml_file).st_size == 0


def random_string(size=8):
    """Generate a random alphanumeric string."""
    chars = (random.SystemRandom().choice(string.ascii_letters)
             for _ in range(size))
    return ''.join(chars)


def generate_password(size=32):
    """Generate a random password using ascii alphabet and digits."""
    chars = (random.SystemRandom().choice(string.ascii_letters + string.digits)
             for _ in range(size))
    return ''.join(chars)


def grep(pattern, file_name):
    """Return lines of a file matching a pattern."""
    return [
        line.rstrip() for line in open(file_name, encoding='utf-8')
        if re.search(pattern, line)
    ]


def gunzip(gzip_file, output_file):
    """Utility to unzip a given gzip file and write it to an output file

    gzip_file: string path to a gzip file
    output_file: string path to the output file
    mode: an octal number to specify file permissions
    """
    output_dir = os.path.dirname(output_file)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, mode=0o755)

    with gzip.open(gzip_file, 'rb') as file_handle:
        contents = file_handle.read()

    def opener(path, flags):
        return os.open(path, flags, mode=0o644)

    with open(output_file, 'wb', opener=opener) as file_handle:
        file_handle.write(contents)


def is_non_empty_file(file_path):
    return os.path.isfile(file_path) and os.path.getsize(file_path) > 0


def is_authenticated_user(username, password):
    """Return true if the user authentication succeeds."""
    import pam  # Minimize dependencies for running tests
    pam_authenticator = pam.pam()
    return bool(pam_authenticator.authenticate(username, password))


class SafeFormatter(string.Formatter):
    """A string.format() handler to deal with missing arguments."""

    def get_value(self, key, args, kwargs):
        """Retrieve a given field's value: 0 or foo."""
        try:
            return super().get_value(key, args, kwargs)
        except (IndexError, KeyError):
            return f'?{key}?'

    def get_field(self, field_name, args, kwargs):
        """Retrieve a given field's value: 0[foo] or foo.bar."""
        try:
            return super().get_field(field_name, args, kwargs)
        except (AttributeError, TypeError):
            return (f'?{field_name}?', '')


def gettext_noop(message: str) -> str:
    """Do nothing. Using this method marks the string for translation."""
    return message



# SPDX-License-Identifier: AGPL-3.0-or-later

from django import http
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import TemplateView
from django.views.generic.edit import FormView

from plinth.modules import first_boot

from .forms import FirstbootWizardSecretForm


class WelcomeView(FormView):
    """Show the welcome screen."""
    form_class = FirstbootWizardSecretForm
    template_name = 'firstboot_welcome.html'

    def form_valid(self, form):
        """If form is valid, mark this step as done and move to next step."""
        self.request.session['firstboot_secret_provided'] = True
        first_boot.mark_step_done('firstboot_welcome')
        return http.HttpResponseRedirect(reverse(first_boot.next_step()))

    def get_context_data(self, **kwargs):
        """Add network connections to context list."""
        context = super().get_context_data(**kwargs)
        show_prompt = first_boot.firstboot_wizard_secret_exists()
        context['show_wizard_password_prompt'] = show_prompt
        return context


class CompleteView(TemplateView):
    """Show next steps after all firstboot wizard steps are done."""

    template_name = 'firstboot_complete.html'

    def get_context_data(self, **kwargs):
        """Add network connections to context list."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Setup Complete')
        return context



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for configure.
"""

from django.utils.translation import gettext_lazy as _

tags = [_('Homepage'), _('Logging'), _('Advanced apps')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for radicale.
"""

import logging

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages, install
from plinth.privileged import service as service_privileged
from plinth.utils import Version, format_lazy

from . import manifest, privileged

_description = [
    format_lazy(
        _('Radicale is a CalDAV and CardDAV server. It allows synchronization '
          'and sharing of scheduling and contact data. To use Radicale, a '
          '<a href="https://radicale.org/master.html#supported-clients">'
          'supported client application</a> is needed. Radicale can '
          'be accessed by any user with a {box_name} login.'),
        box_name=_(cfg.box_name)),
    _('Radicale provides a basic web interface, which only supports creating '
      'new calendars and addressbooks. It does not support adding events or '
      'contacts, which must be done using a separate client.'),
]

logger = logging.getLogger(__name__)


class RadicaleApp(app_module.App):
    """FreedomBox app for Radicale."""

    app_id = 'radicale'

    _version = 6

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Radicale'), icon_filename='radicale',
                               description=_description,
                               manual_page='Radicale',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-radicale', info.name, info.icon_filename,
                              info.tags, 'radicale:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-radicale', info.name,
                                      icon=info.icon_filename,
                                      url='/radicale/', clients=info.clients,
                                      tags=info.tags, login_required=True)
        self.add(shortcut)

        packages = Packages('packages-radicale', ['radicale'],
                            rerun_setup_on_upgrade=True)
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-radicale', [
            '/etc/apache2/conf-available/radicale2-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-radicale', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-radicale', 'radicale2-freedombox',
                              urls=['https://{host}/radicale'])
        self.add(webserver)

        daemon = Daemon('daemon-radicale', 'uwsgi-app@radicale.socket')
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-radicale',
                                          reserved_usernames=['radicale'])
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-radicale',
                                       **manifest.backup)
        self.add(backup_restore)

    def enable(self):
        """Fix missing directories before enabling radicale."""
        privileged.fix_paths()
        super().enable()

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

        if old_version and old_version <= 4:
            webserver = self.get_component('webserver-radicale')
            daemon = self.get_component('daemon-radicale')
            if webserver.is_enabled():
                daemon.enable()

            # Vanquish the old uwsgi init.d script.
            service_privileged.disable('uwsgi')
            service_privileged.mask('uwsgi')

    def force_upgrade(self, packages):
        """Force upgrade radicale to resolve conffile prompt."""
        if 'radicale' not in packages:
            return False

        # Allow upgrade from 3.1.8 (bookworm) to 3.4.1 (trixie) and beyond 3.x.
        package = packages['radicale']
        if Version(package['new_version']) > Version('4~'):
            return False

        rights = privileged.get_rights_value()
        install(['radicale'], force_configuration='new')
        privileged.setup()
        privileged.configure(rights)

        return True

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for transmission app.
"""

import pytest
from selenium.webdriver.common.keys import Keys

from plinth.tests import functional

pytestmark = [pytest.mark.tags]


def _is_app_listed(session_browser, app):
    """Assert that the specified app is listed on the page."""
    app_links = session_browser.links.find_by_href(f'/freedombox/apps/{app}/')
    assert len(app_links) == 1


@pytest.fixture(name='bittorrent_tag')
def fixture_bittorrent_tag(session_browser):
    """Click on the BitTorrent tag."""
    bittorrent_tag = '/freedombox/apps/?tag=BitTorrent'
    functional.login(session_browser)
    functional.nav_to_module(session_browser, 'transmission')
    with functional.wait_for_page_update(session_browser, timeout=10,
                                         expected_url=bittorrent_tag):
        session_browser.links.find_by_href(bittorrent_tag).click()


@pytest.fixture(name='locale')
def fixture_locale(session_browser):
    """Set a different language for a user."""
    functional.login(session_browser)
    try:
        functional.user_set_language(session_browser, 'es')
        yield
    finally:
        functional.user_set_language(session_browser, '')


def test_bittorrent_tag(session_browser, bittorrent_tag):
    """Test that the BitTorrent tag lists Deluge and Transmission."""
    _is_app_listed(session_browser, 'deluge')
    _is_app_listed(session_browser, 'transmission')


def test_search_for_tag(session_browser, bittorrent_tag):
    """Test that searching for a tag returns the expected apps."""
    search_input = session_browser.find_by_id('add-tag-input').first
    with functional.wait_for_page_update(
            session_browser, timeout=10,
            expected_url='/freedombox/apps/?tag=BitTorrent&tag=File+sharing'):
        search_input.click()
        search_input.type('file sharing')
        search_input.type(Keys.ENTER)

    for app in ['deluge', 'samba', 'sharing', 'syncthing', 'transmission']:
        _is_app_listed(session_browser, app)


def test_click_on_tag(session_browser, bittorrent_tag):
    """Test that clicking on a tag lists the expected apps."""
    search_input = session_browser.find_by_id('add-tag-input').first
    with functional.wait_for_page_update(
            session_browser, timeout=10,
            expected_url='/freedombox/apps/?tag=BitTorrent&tag=File+sync'):
        search_input.click()
        session_browser.find_by_css(
            ".dropdown-item[data-tag='File sync']").click()

    for app in ['deluge', 'nextcloud', 'syncthing', 'transmission']:
        _is_app_listed(session_browser, app)


def test_tag_localization(session_browser, locale):
    """Test that tags are localized and tests in done localized."""
    functional.visit(session_browser, '/freedombox/apps/?tag=Sharing')
    badge = session_browser.find_by_css('.tag[data-tag="Sharing"]').first
    assert 'Compartir' in badge.text

    search_input = session_browser.find_by_id('add-tag-input').first
    with functional.wait_for_page_update(
            session_browser, timeout=10,
            expected_url='/freedombox/apps/?tag=Sharing&tag=Bookmarks'):
        search_input.click()
        search_input.type('Marcadores')
        search_input.type(Keys.ENTER)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Application manifest for privacy app."""

from django.utils.translation import gettext_lazy as _

from . import privileged

backup = {'config': {'files': [str(privileged.CONFIG_FILE)]}}

tags = [_('Usage reporting'), _('External services'), _('Fallback DNS')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Quassel."""

import pathlib

from plinth.actions import privileged


@privileged
def set_domain(domain_name: str):
    """Write a file containing domain name."""
    domain_file = pathlib.Path('/var/lib/quassel/domain-freedombox')
    domain_file.write_text(domain_name, encoding='utf-8')
    # Ensure that that file is readable by non-privileged process.
    domain_file.chmod(0o644)



"""Custom OpenID Connect validators."""

import ipaddress
import urllib.parse

from oauth2_provider import oauth2_validators

from plinth import action_utils


class OAuth2Validator(oauth2_validators.OAuth2Validator):
    """Add/override claims into Discovery and ID token.

    Ensure that basic profile information is available to the clients. Make the
    value of the 'sub' claim, as defined in OpenID Connect, to be the username
    of the account instead of the Django account ID. The username is unique in
    FreedomBox.

    We wish for the applications using the Identity Provider to also
    provide/deny resources based on the groups that the user is part of. For
    this, we add an additional scope "freedombox_groups" and additional claim
    "freedombox_groups". To define custom scopes and claims, we need to ensure
    that the keys used are unique and will not clash with other
    implementations. 'freedombox_' prefix seems reasonable. The value of this
    claim is a list of all groups that the user account is part of.
    """

    # Add a scope, as recommended in the oauth-toolkit documentation.
    oidc_claim_scope = oauth2_validators.OAuth2Validator.oidc_claim_scope
    oidc_claim_scope.update({'freedombox_groups': 'freedombox_groups'})

    def get_additional_claims(self):
        """Override value of 'sub' claim and add other claims.

        Only the 'sub' claim is filled by default by django-oauth-toolkit. The
        rest, as needed, must be filled by us.

        Use the 'second' form of get_additional_claims override as
        documentation suggests so that the base code and automatically add the
        list of claims returned here to list of supported claims.
        """

        def _get_user_groups(request):
            return list(request.user.groups.values_list('name', flat=True))

        return {
            'sub': lambda request: request.user.username,
            'email': lambda request: request.user.email,
            'preferred_username': lambda request: request.user.username,
            'freedombox_groups': _get_user_groups,
        }

    def validate_redirect_uri(self, client_id, redirect_uri, request, *args,
                              **kwargs):
        """Additionally allow redirection to this server's IPs and domains."""
        allowed_redirect_uris = request.client.redirect_uris.split()
        if _validate_local_domains_and_ips(redirect_uri, request,
                                           allowed_redirect_uris):
            return True

        return super().validate_redirect_uri(client_id, redirect_uri, request,
                                             *args, **kwargs)


def _validate_local_domains_and_ips(redirect_uri, request,
                                    allowed_redirect_uris):
    """Allow redirect to local domains and IPs.

    See models.py:redirect_to_uri_allowed() in django-oauth-toolkit for
    reference.

    Path and query part of the redirect URL must always match with one of the
    configured redirect URLs for this application. The query in the redirect
    URI is allowed to be a subset of the allowed query.

    Localhost domains (localhost, ip6-localhost, and ip6-loopback) are allowed
    in redirect URLs. Scheme and port are not checked.

    An IP address based redirect URL is accepted as long as it is to the same
    IP address with which the FreedomBox's Identity Provider is being accessed.
    Scheme is not checked. Changing IP address during OpenID Connect flow is
    not allowed.
    """
    # Requires 'ProxyPreserveHost On' in Apache2 configuration for proxying
    # requests to FreedomBox service.
    request_host = request.headers.get('HTTP_HOST')

    parsed_redirect_uri = urllib.parse.urlparse(redirect_uri)

    redirect_uri_query_set = set(
        urllib.parse.parse_qs(parsed_redirect_uri.query))
    try:
        ipaddress.ip_address(parsed_redirect_uri.hostname)
        redirect_uri_is_ip = True
    except ValueError:
        redirect_uri_is_ip = False

    redirect_uri_is_localhost = parsed_redirect_uri.hostname in (
        'localhost', 'ip6-localhost', 'ip6-loopback',
        action_utils.get_hostname())

    for allowed_uri in allowed_redirect_uris:
        parsed_allowed_uri = urllib.parse.urlparse(allowed_uri)

        # Path must match one of the allowed paths
        if parsed_redirect_uri.path != parsed_allowed_uri.path:
            continue

        # Query must be a subset of allowed query
        allowed_query_set = set(urllib.parse.parse_qs(
            parsed_allowed_uri.query))
        if not allowed_query_set.issubset(redirect_uri_query_set):
            continue

        # If the redirect is to an IP address, it is only allowed if the IDP
        # itself is being accessed with that IP address.
        if (redirect_uri_is_ip and request_host
                and parsed_redirect_uri.netloc == request_host):
            return True

        # If the redirect is to a 'localhost' like address, a port mismatch is
        # allowed.
        if redirect_uri_is_localhost:
            return True

    return False  # Special criteria didn't match, do usual checks.



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the Home Assistant module."""

from django.urls import re_path

from .views import HomeAssistantAppView

urlpatterns = [
    re_path(r'^apps/homeassistant/$', HomeAssistantAppView.as_view(),
            name='index')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for minetest module."""

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth.modules import names
from plinth.views import AppView

from . import get_configuration, privileged
from .forms import MinetestForm


class MinetestAppView(AppView):  # pylint: disable=too-many-ancestors
    """A specialized view for configuring minetest."""

    app_id = 'minetest'
    template_name = 'minetest.html'
    form_class = MinetestForm

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial.update(get_configuration())
        return initial

    def get_context_data(self, *args, **kwargs) -> dict[str, object]:
        """Add service to the context data."""
        context = super().get_context_data(*args, **kwargs)
        context['domains'] = names.components.DomainName.list_names(
            'minetest-plinth')
        return context

    def form_valid(self, form):
        """Change the configurations of Minetest service."""
        data = form.cleaned_data
        old_config = get_configuration()

        changes = {}
        if old_config['max_players'] != data['max_players'] \
           and data['max_players'] is not None:
            changes['max_players'] = data['max_players']

        if old_config['creative_mode'] != data['creative_mode']:
            changes['creative_mode'] = data['creative_mode']

        if old_config['enable_pvp'] != data['enable_pvp']:
            changes['enable_pvp'] = data['enable_pvp']

        if old_config['enable_damage'] != data['enable_damage']:
            changes['enable_damage'] = data['enable_damage']

        if changes:
            privileged.configure(**changes)
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for storage module operations.
"""

import contextlib
import dataclasses
import re
import subprocess
import tempfile
from unittest.mock import patch

import psutil
import pytest
from django.utils.translation import gettext_noop

from plinth.diagnostic_check import DiagnosticCheck, Result
from plinth.modules import storage
from plinth.modules.storage import privileged


def _is_container():
    """Return if we are running inside a container."""
    try:
        process = subprocess.run(['systemd-detect-virt', '--container'],
                                 check=False)
    except FileNotFoundError:
        # systemd is not installed on the machine. Likely a container.
        return True  # Skip just in case.

    return process.returncode == 0


pytestmark = pytest.mark.usefixtures('mock_privileged', 'mock_run_as_user')
privileged_modules_to_mock = ['plinth.modules.storage.privileged']
skip_if_container = pytest.mark.skipif(_is_container(),
                                       reason='running inside a container')


class Disk():
    """Context manager to create/destroy a disk."""

    def __init__(self, size, disk_info, file_system_info=None):
        """Initialize the context manager object."""
        self.size = size
        self.disk_info = disk_info
        self.file_system_info = file_system_info

        self.disk_file = None
        self.device = None

    def _create_disk_file(self):
        """Create a temporary file to act as a disk."""
        disk_file = tempfile.NamedTemporaryFile()

        command = 'dd if=/dev/zero of={file} bs=1M count={size}' \
                  .format(size=self.size, file=disk_file.name)
        subprocess.run(command.split(), stderr=subprocess.DEVNULL, check=True)

        self.disk_file = disk_file

    def expand_disk_file(self, size):
        """Expand the disk file."""
        command = f'truncate --size={size}M {self.disk_file.name}'
        subprocess.run(command.split(), check=True)
        self._unmount_file_systems()
        self._cleanup_loopback()
        self._setup_loopback()

    def get_partition_device(self, partition_number):
        """Return the device corresponding to a partition in a given device."""
        if re.match('[0-9]', self.device[-1]):
            return self.device + 'p' + str(partition_number)

        return self.device + str(partition_number)

    @contextlib.contextmanager
    def mount_partition(self, partition_number):
        """Mount partition onto a directory if device has a filesystem."""
        device = self.get_partition_device(partition_number)
        if not self.file_system_info or \
           partition_number not in dict(self.file_system_info):
            yield '/'
            return

        with tempfile.TemporaryDirectory() as mount_path:
            subprocess.run(['mount', device, mount_path], check=True)
            try:
                yield mount_path
            finally:
                subprocess.run(['umount', mount_path], check=True)

    def _setup_loopback(self):
        """Setup loop back on the create disk file."""
        command = 'losetup --show --find {file}'.format(
            file=self.disk_file.name)
        process = subprocess.run(command.split(), stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE, check=False)
        if process.returncode:
            if b'cannot find an unused loop device' in process.stderr:
                pytest.skip('Loopback devices not available')
            else:
                raise Exception(process.stderr)

        device = process.stdout.decode().strip()

        subprocess.run(['partprobe', device], check=True)

        self.device = device

    def _create_partitions(self):
        """Create partitions as specified in disk_info."""
        steps = [step.split() for step in self.disk_info]
        command = [
            'parted', '--align=optimal', '--script', self.disk_file.name
        ]
        for step in steps:
            command += step

        subprocess.run(command, check=True)

    def _create_file_systems(self):
        """Create file systems inside partitions."""
        if not self.file_system_info:
            return

        for partition, file_system_type in self.file_system_info:
            device = self.get_partition_device(partition)
            if file_system_type == 'btrfs':
                command = ['mkfs.btrfs', '-K', device]
            else:
                command = ['mkfs.ext4', device]

            subprocess.run(command, stdout=subprocess.DEVNULL,
                           stderr=subprocess.DEVNULL, check=True)

    def _unmount_file_systems(self):
        """Unmount al partitions if it is mounted by external party."""
        if not self.file_system_info:
            return

        for partition, _ in self.file_system_info:
            device = self.get_partition_device(partition)
            subprocess.run(['umount', device], check=False)

    def _cleanup_loopback(self):
        """Undo the loopback device setup."""
        subprocess.run(['losetup', '--detach', self.device], check=False)

    def _remove_disk_file(self):
        """Delete the disk_file."""
        self.disk_file.close()
        self.disk_file = None

    def __enter__(self):
        """Enter the context, create the test disk."""
        self._create_disk_file()
        self._create_partitions()
        self._setup_loopback()
        self._create_file_systems()
        return self

    def __exit__(self, *exc):
        """Exit the context, destroy the test disk."""
        self._unmount_file_systems()
        self._cleanup_loopback()
        self._remove_disk_file()


def disk_space_available():
    """Return disk space available in temporary directory."""
    directory = tempfile.gettempdir()
    output = subprocess.check_output(['df', '-m', '--output=avail',
                                      directory]).decode()
    return int(output.splitlines()[1])


@pytest.mark.usefixtures('needs_root')
@pytest.mark.skipif(disk_space_available() < 1024,
                    reason='Needs 1024MiB of space')
@skip_if_container
def test_simple_case():
    """Test a simple with no complications"""
    disk_info = [
        'mktable msdos', 'mkpart primary btrfs 1 300',
        'mkpart primary btrfs 301 600', 'mkpart primary btrfs 608 900'
    ]
    # Btrfs volumes < 256 MiB can't be resized.
    # https://bugzilla.kernel.org/show_bug.cgi?id=118111
    with Disk(1024, disk_info, [(1, 'btrfs'), (2, 'btrfs'),
                                (3, 'btrfs')]) as disk:
        # No free space
        _assert_free_space(disk, 1, space=False)
        # < 10 MiB of free space
        _assert_free_space(disk, 2, space=False)
        _assert_free_space(disk, 3, space=True)

        _expand_partition(disk, 1, success=False)
        _expand_partition(disk, 2, success=False)
        _expand_partition(disk, 3, success=True)
        _expand_partition(disk, 3, success=False)


@pytest.mark.usefixtures('needs_root')
@pytest.mark.skipif(disk_space_available() < 512,
                    reason='Needs 512MiB of space')
@skip_if_container
def test_extended_partition_free_space():
    """Test that free space does not show up when outside extended."""
    disk_info = [
        'mktable msdos', 'mkpart primary 1 8', 'mkpart extended 8 500',
        'mkpart logical btrfs 9 300'
    ]
    with Disk(512, disk_info, [(5, 'btrfs')]) as disk:
        _assert_free_space(disk, 5, space=False)
        _expand_partition(disk, 5, success=False)


@pytest.mark.usefixtures('needs_root')
@pytest.mark.skipif(disk_space_available() < 512,
                    reason='Needs 512MiB of space')
@skip_if_container
def test_gpt_partition_free_space():
    """Test that GPT partitions can be expanded."""
    # Specifically check for partition number > 4
    disk_info = [
        'mktable gpt', 'mkpart primary 1 4', 'mkpart extended 4 8',
        'mkpart extended 8 12', 'mkpart extended 12 16',
        'mkpart extended 16 300'
    ]
    with Disk(512, disk_info, [(5, 'btrfs')]) as disk:
        # Second header already at the end
        _assert_free_space(disk, 5, space=True)
        _expand_partition(disk, 5, success=True)
        _expand_partition(disk, 5, success=False)
        disk.expand_disk_file(1024)
        # Second header not at the end
        _assert_free_space(disk, 5, space=True)
        _expand_partition(disk, 5, success=True)
        _expand_partition(disk, 5, success=False)


@pytest.mark.usefixtures('needs_root')
@pytest.mark.parametrize('partition_table_type', ['gpt', 'msdos'])
@pytest.mark.skipif(disk_space_available() < 32, reason='Needs 32MiB of space')
@skip_if_container
def test_unsupported_file_system(partition_table_type):
    """Test that free space after unknown file system does not count."""
    disk_info = [f'mktable {partition_table_type}', 'mkpart primary 1 8']
    with Disk(32, disk_info) as disk:
        _assert_free_space(disk, 1, space=False)
        _expand_partition(disk, 1, success=False)


@pytest.mark.usefixtures('needs_root')
@pytest.mark.parametrize('partition_table_type', ['gpt', 'msdos'])
@pytest.mark.skipif(disk_space_available() < 512,
                    reason='Needs 512MiB of space')
@skip_if_container
def test_btrfs_expansion(partition_table_type):
    """Test that btrfs file system can be expanded."""
    disk_info = [
        f'mktable {partition_table_type}', 'mkpart primary btrfs 1 300'
    ]
    with Disk(512, disk_info, [(1, 'btrfs')]) as disk:
        _expand_partition(disk, 1, success=True)
        _expand_partition(disk, 1, success=False)
        _assert_btrfs_file_system_healthy(disk, 1)


@pytest.mark.usefixtures('needs_root')
@pytest.mark.parametrize('partition_table_type', ['gpt', 'msdos'])
@pytest.mark.skipif(disk_space_available() < 128,
                    reason='Needs 128MiB of space')
@skip_if_container
def test_ext4_expansion(partition_table_type):
    """Test that ext4 file system can be expanded."""
    disk_info = [f'mktable {partition_table_type}', 'mkpart primary ext4 1 64']
    with Disk(128, disk_info, [(1, 'ext4')]) as disk:
        _expand_partition(disk, 1, success=True)
        _expand_partition(disk, 1, success=False)
        _assert_ext4_file_system_healthy(disk, 1)


def _assert_free_space(disk, partition_number, space=True):
    """Verify that free is available/not available after a partition."""
    device = disk.get_partition_device(partition_number)
    if space:
        privileged.is_partition_expandable(device)
    else:
        with pytest.raises(RuntimeError):
            privileged.is_partition_expandable(device)


def _expand_partition(disk, partition_number, success=True):
    """Expand a partition."""
    _assert_aligned(disk, partition_number)
    with disk.mount_partition(partition_number) as mount_point:
        device = disk.get_partition_device(partition_number)
        if success:
            privileged.expand_partition(device, mount_point)
        else:
            with pytest.raises(RuntimeError):
                privileged.expand_partition(device, mount_point)

    _assert_aligned(disk, partition_number)


def _assert_aligned(disk, partition_number):
    """Test that partition is optimally aligned."""
    subprocess.run([
        'parted', '--script', disk.device, 'align-check', 'opti',
        str(partition_number)
    ], check=True)


def _assert_btrfs_file_system_healthy(disk, partition_number):
    """Perform a successful ext4 file system check."""
    device = disk.get_partition_device(partition_number)
    command = ['btrfs', 'check', '--force', device]
    subprocess.run(command, stdout=subprocess.DEVNULL,
                   stderr=subprocess.DEVNULL, check=True)


def _assert_ext4_file_system_healthy(disk, partition_number):
    """Perform a successful ext4 file system check."""
    device = disk.get_partition_device(partition_number)
    command = ['e2fsck', '-n', device]
    subprocess.run(command, stdout=subprocess.DEVNULL,
                   stderr=subprocess.DEVNULL, check=True)


def _assert_validate_directory(path, error, check_writable=False,
                               check_creatable=False):
    """Perform directory validation checks."""
    if error:
        match = None if not error.args else error.args[0]
        with pytest.raises(error.__class__, match=match):
            privileged.validate_directory(path, check_creatable,
                                          check_writable,
                                          for_user='test-non-existent')
    else:
        privileged.validate_directory(path, check_creatable, check_writable,
                                      for_user='test-non-existent')


@pytest.mark.usefixtures('needs_not_root')
@pytest.mark.parametrize('path,error',
                         [('/missing', FileNotFoundError()),
                          ('/etc/os-release', NotADirectoryError()),
                          ('/root', PermissionError('read')), ('/', None),
                          ('/etc/..', None)])
def test_validate_directory(path, error):
    """Test that directory validation returns expected output."""
    _assert_validate_directory(path, error)


@pytest.mark.usefixtures('needs_not_root')
@pytest.mark.parametrize('path,error', [('/', PermissionError('write')),
                                        ('/tmp', None)])
def test_validate_directory_writable(path, error):
    """Test that directory writable validation returns expected output."""
    _assert_validate_directory(path, error, check_writable=True)


@pytest.mark.usefixtures('needs_not_root')
@pytest.mark.parametrize(
    'path,error',
    [('/var/lib/plinth_storage_test_not_exists', PermissionError('write')),
     ('/tmp/plint_storage_test_not_exists', None),
     ('/var/../tmp/plint_storage_test_not_exists', None)])
def test_validate_directory_creatable(path, error):
    """Test that directory creatable validation returns expected output."""
    _assert_validate_directory(path, error, check_creatable=True)


@patch('psutil.disk_partitions')
def test_is_partition_read_only(disk_partitions):
    """Test whether checking for ro partition works."""

    def partition(*args):
        try:
            # psutil <= 5.9.8
            # pylint: disable=protected-access
            return psutil._common.sdiskpart(*args)
        except TypeError:
            # psutil >= 7.0
            return psutil._common.sdiskpart(*args[:-2])

    disk_partitions.return_value = [
        partition('/dev/root', '/', 'btrfs', 'rw,nosuid', 42, 42),
        partition('/dev/root', '/foo', 'btrfs', 'rw', 321, 321),
        partition('/dev/foo', '/bar', 'extfs', 'ro', 123, 123)
    ]
    assert not storage.is_partition_read_only('/')
    assert not storage.is_partition_read_only('/foo')
    assert storage.is_partition_read_only('/bar')


@patch('subprocess.check_output')
def test_diagnose_grub_configured(check_output):
    """Test whether checking grub-pc package configuration failed works."""
    diagnose = storage._diagnose_grub_configured \
        # pylint: disable=protected-access

    passed = DiagnosticCheck('storage-grub-configured',
                             gettext_noop('grub package is configured'),
                             Result.PASSED)
    failed = dataclasses.replace(passed, result=Result.FAILED)
    warning = dataclasses.replace(passed, result=Result.WARNING)

    # installed and configured
    check_output.return_value = b'ii '
    assert diagnose() == passed

    # failed configuration
    check_output.return_value = b'iF '
    assert diagnose() == failed

    # should be installed, but somehow is not
    check_output.return_value = b'in '
    assert diagnose() == warning

    # not installed
    check_output.return_value = b'un '
    assert diagnose() is None

    # grub-pc package is not available (e.g. ARM devices)
    check_output.side_effect = subprocess.CalledProcessError(
        cmd=[
            'dpkg-query', '--show', '--showformat=${db:Status-Abbrev}',
            'grub-pc'
        ], returncode=1,
        stderr=b'dpkg-query: no packages found matching grub-pc\n')
    assert diagnose() is None



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test various part of the dist upgrade process.
"""

import re
import subprocess
from datetime import datetime as datetime_original
from datetime import timezone
from unittest.mock import Mock, call, patch

import pytest

from plinth.modules.upgrades import distupgrade

# pylint: disable=protected-access


@patch('subprocess.run')
def test_apt_run(run):
    """Test that running apt command logs properly."""
    run.return_value.returncode = 0
    args = ['command', 'arg1', 'arg2']
    distupgrade._apt_run(args)
    assert run.call_args.args == \
        (['apt-get', '--assume-yes', '--quiet=2'] + args,)
    assert run.call_args.kwargs['stdout'] == subprocess.PIPE

    run.return_value.returncode = 10
    with pytest.raises(RuntimeError):
        distupgrade._apt_run(args)


def test_sources_list_update(tmp_path):
    """Test that updating a sources file works."""
    original = '''
# This is a comment with 'bookworm' in it.
deb http://deb.debian.org/debian bookworm main non-free-firmware
deb-src http://deb.debian.org/debian bookworm main non-free-firmware

deb http://deb.debian.org/debian bookworm-updates main non-free-firmware
deb-src http://deb.debian.org/debian bookworm-updates main non-free-firmware

deb http://security.debian.org/debian-security/ bookworm-security main non-free-firmware
deb-src http://security.debian.org/debian-security/ bookworm-security main non-free-firmware

deb https://deb.debian.org/debian other main
deb https://deb.debian.org/debian bookwormish main
'''  # noqa: E501
    modified = '''
# This is a comment with 'bookworm' in it.
deb http://deb.debian.org/debian trixie main non-free-firmware
deb-src http://deb.debian.org/debian trixie main non-free-firmware

deb http://deb.debian.org/debian trixie-updates main non-free-firmware
deb-src http://deb.debian.org/debian trixie-updates main non-free-firmware

deb http://security.debian.org/debian-security/ trixie-security main non-free-firmware
deb-src http://security.debian.org/debian-security/ trixie-security main non-free-firmware

deb https://deb.debian.org/debian other main
deb https://deb.debian.org/debian bookwormish main
'''  # noqa: E501

    sources_list = tmp_path / 'sources.list'
    temp_sources_list = tmp_path / 'sources.list.fbx-dist-upgrade'

    module = 'plinth.modules.upgrades.distupgrade'
    with patch(f'{module}.sources_list', sources_list), \
         patch(f'{module}.temp_sources_list', temp_sources_list):
        sources_list.write_text(original)
        distupgrade._sources_list_update('bookworm', 'trixie')

        assert temp_sources_list.read_text() == modified

        original = re.sub(r'bookworm([ -])', r'stable\1', original)
        sources_list.write_text(original)
        distupgrade._sources_list_update('bookworm', 'trixie')

        assert temp_sources_list.read_text() == modified


@patch('datetime.datetime')
@patch('plinth.modules.upgrades.utils.get_current_release')
@patch('plinth.modules.upgrades.distupgrade.utils.get_sources_list_codename')
@patch('plinth.action_utils.service_is_running')
@patch('plinth.modules.upgrades.utils.is_sufficient_free_space')
@patch('plinth.modules.upgrades.is_dist_upgrade_enabled')
@patch('plinth.modules.upgrades.utils.check_auto')
def test_get_status(check_auto, is_dist_upgrade_enabled,
                    is_sufficient_free_space, service_is_running,
                    get_sources_list_codename, get_current_release, datetime):
    """Test getting status of distribution upgrade."""
    # All checks fail, sources.list has 'testing' value
    check_auto.return_value = False
    is_dist_upgrade_enabled.return_value = False
    is_sufficient_free_space.return_value = False
    service_is_running.return_value = True
    get_sources_list_codename.return_value = 'testing'
    status = distupgrade.get_status()
    assert not status['updates_enabled']
    assert not status['dist_upgrade_enabled']
    assert not status['has_free_space']
    assert status['running']
    assert status['current_codename'] == 'testing'
    assert status['current_version'] is None
    assert status['current_release_date'] is None
    assert status['next_codename'] is None
    assert status['next_version'] is None
    assert status['next_release_date'] is None
    assert status['next_action'] is None

    # sources.list has mixed values
    get_sources_list_codename.return_value = None
    status = distupgrade.get_status()
    assert status['current_codename'] is None
    assert status['current_version'] is None

    # sources.list has 'unstable' value
    get_sources_list_codename.return_value = 'unstable'
    status = distupgrade.get_status()
    assert status['current_codename'] == 'unstable'
    assert status['current_version'] is None

    # sources.list has an unknown value
    get_sources_list_codename.return_value = 'x-invalid'
    get_current_release.return_value = (None, None)
    status = distupgrade.get_status()
    assert status['current_codename'] == 'x-invalid'
    assert status['current_version'] is None

    # sources.list has 'stable'
    get_sources_list_codename.return_value = 'stable'
    get_current_release.return_value = (None, 'bookworm')
    status = distupgrade.get_status()
    assert status['current_codename'] == 'bookworm'
    assert status['current_version'] == 12

    # All checks pass, next release not yet available
    check_auto.return_value = True
    is_dist_upgrade_enabled.return_value = True
    is_sufficient_free_space.return_value = True
    service_is_running.return_value = False
    get_sources_list_codename.return_value = 'bookworm'
    get_current_release.return_value = (None, 'bookworm')
    datetime.now.return_value = datetime_original(2024, 8, 10,
                                                  tzinfo=timezone.utc)
    status = distupgrade.get_status()
    assert status['updates_enabled']
    assert status['dist_upgrade_enabled']
    assert status['has_free_space']
    assert not status['running']
    assert status['current_codename'] == 'bookworm'
    assert status['current_version'] == 12
    current_date = datetime_original(2023, 6, 10, tzinfo=timezone.utc)
    assert status['current_release_date'] == current_date
    assert status['next_codename'] == 'trixie'
    assert status['next_version'] == 13
    next_date = datetime_original(2025, 8, 9, tzinfo=timezone.utc)
    assert status['next_release_date'] == next_date
    assert status['next_action'] == 'manual'

    # Distribution upgrade interrupted
    get_current_release.return_value = (None, 'trixie')
    status = distupgrade.get_status()
    assert status['next_action'] == 'continue'

    # Less than 30 days after release
    get_current_release.return_value = (None, 'bookworm')
    datetime.now.return_value = datetime_original(2025, 8, 30,
                                                  tzinfo=timezone.utc)
    status = distupgrade.get_status()
    assert status['next_action'] == 'wait_or_manual'

    # More than 30 days after release
    datetime.now.return_value = datetime_original(2025, 9, 30,
                                                  tzinfo=timezone.utc)
    status = distupgrade.get_status()
    assert status['next_action'] == 'ready'

    # Next release date not available
    get_sources_list_codename.return_value = 'trixie'
    assert distupgrade.get_status()['next_action'] is None

    # Automatic updates not enabled
    get_sources_list_codename.return_value = 'bookworm'
    check_auto.return_value = False
    assert distupgrade.get_status()['next_action'] is None

    # Distribution updates not enabled
    check_auto.return_value = True
    is_dist_upgrade_enabled.return_value = False
    assert distupgrade.get_status()['next_action'] is None

    # Not enough free space
    is_dist_upgrade_enabled.return_value = True
    is_sufficient_free_space.return_value = False
    assert distupgrade.get_status()['next_action'] is None

    # Distribution upgrade running
    is_sufficient_free_space.return_value = True
    service_is_running.return_value = True
    assert distupgrade.get_status()['next_action'] is None


@patch('subprocess.run')
@patch('plinth.modules.snapshot.is_apt_snapshots_enabled')
@patch('plinth.modules.snapshot.is_supported')
def test_snapshot_run_and_disable(is_supported, is_apt_snapshots_enabled, run):
    """Test taking a snapshot."""
    is_supported.return_value = False
    with distupgrade._snapshot_run_and_disable():
        run.assert_not_called()

    run.assert_not_called()

    is_supported.return_value = True
    is_apt_snapshots_enabled.return_value = False
    with distupgrade._snapshot_run_and_disable():
        assert run.call_args_list == [
            call(['snapper', 'create', '--description', 'before dist-upgrade'],
                 stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        ]
        run.reset_mock()

    run.assert_not_called()

    is_supported.return_value = True
    is_apt_snapshots_enabled.return_value = True
    with distupgrade._snapshot_run_and_disable():
        assert run.call_args_list == [
            call(['snapper', 'create', '--description', 'before dist-upgrade'],
                 stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True),
            call([
                '/usr/bin/freedombox-cmd', 'snapshot', 'disable_apt_snapshot'
            ], input=b'{"args": ["yes"], "kwargs": {}}',
                 stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        ]
        run.reset_mock()

    assert run.call_args_list == [
        call(['/usr/bin/freedombox-cmd', 'snapshot', 'disable_apt_snapshot'],
             input=b'{"args": ["no"], "kwargs": {}}', stdout=subprocess.PIPE,
             stderr=subprocess.PIPE, check=True)
    ]


@patch('plinth.action_utils.service_enable')
@patch('plinth.action_utils.service_disable')
@patch('plinth.action_utils.service_is_running')
def test_services_disable(service_is_running, service_disable, service_enable):
    """Test that disabling services works."""
    service_is_running.return_value = False
    with distupgrade._services_disable():
        service_disable.assert_not_called()

    service_enable.assert_not_called()

    service_is_running.return_value = True
    with distupgrade._services_disable():
        service_disable.call_args_list = [call('quasselcore')]

    service_enable.call_args_list = [call('quasselcore')]


@patch('subprocess.run')
def test_apt_hold_packages(run, tmp_path):
    """Test that holding apt packages works."""

    def _run(command, **kwargs):
        if 'showhold' in command:
            return Mock(stdout=False)

        return Mock(returncode=0)

    hold_flag = tmp_path / 'flag'
    run.side_effect = _run
    with patch('plinth.action_utils.apt_hold_flag', hold_flag), \
         patch('plinth.modules.upgrades.distupgrade.PACKAGES_WITH_PROMPTS',
               ['package1', 'package2']):
        with distupgrade._apt_hold_packages():
            assert hold_flag.exists()
            assert hold_flag.stat().st_mode & 0o117 == 0
            expected_calls = [
                call(['apt-mark', 'showhold', 'freedombox'], check=True,
                     stdout=subprocess.PIPE, stderr=subprocess.PIPE),
                call(['apt-mark', 'hold', 'freedombox'], check=True,
                     stdout=subprocess.PIPE, stderr=subprocess.PIPE),
                call(['apt-mark', 'showhold', 'package1'],
                     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                     check=True),
                call(['apt-mark', 'hold', 'package1'], check=False,
                     stdout=subprocess.PIPE, stderr=subprocess.PIPE),
                call(['apt-mark', 'showhold', 'package2'],
                     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                     check=True),
                call(['apt-mark', 'hold', 'package2'], check=False,
                     stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            ]
            assert run.mock_calls == expected_calls
            run.reset_mock()

        expected_call = [
            call(['apt-mark', 'unhold', 'package1'], stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE, check=True),
            call(['apt-mark', 'unhold', 'package2'], stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE, check=True),
            call(['apt-mark', 'unhold', 'freedombox'], stdout=subprocess.PIPE,
                 stderr=subprocess.PIPE, check=False),
        ]
        assert run.call_args_list == expected_call


@patch('plinth.action_utils.debconf_set_selections')
def test_debconf_set_selections(debconf_set_selections):
    """Test that setting debconf selections works."""
    selections = 'plinth.modules.upgrades.distupgrade.PRE_DEBCONF_SELECTIONS'
    with patch(selections, []):
        distupgrade._debconf_set_selections()
        debconf_set_selections.assert_not_called()

    with patch(selections, ['selection1', 'selection2']):
        distupgrade._debconf_set_selections()
        debconf_set_selections.assert_called_with(['selection1', 'selection2'])

    distupgrade._debconf_set_selections()
    debconf_set_selections.assert_called_with(
        ['grub-pc grub-pc/install_devices_empty boolean true'])


@patch('plinth.modules.upgrades.distupgrade._apt_run')
def test_packages_remove_obsolete(apt_run):
    """Test that obsolete packages are removed."""
    distupgrade._packages_remove_obsolete()
    apt_run.assert_not_called()  # No obsolete package to remove currently.

    with patch('plinth.modules.upgrades.distupgrade.OBSOLETE_PACKAGES',
               ['searx']):
        distupgrade._packages_remove_obsolete()
        apt_run.assert_called_with(['remove', 'searx'])


@patch('plinth.modules.upgrades.distupgrade._apt_run')
def test_apt_update(apt_run):
    """Test that apt update works."""
    distupgrade._apt_update()
    apt_run.assert_called_with(['update'])


@patch('plinth.modules.upgrades.distupgrade._apt_run')
@patch('subprocess.run')
def test_apt_fix(run, apt_run):
    """Test that apt fixes work."""
    distupgrade._apt_fix()
    assert run.call_args_list == [
        call(['dpkg', '--configure', '-a'], stdout=subprocess.PIPE,
             stderr=subprocess.PIPE, check=False)
    ]
    assert apt_run.call_args_list == [call(['--fix-broken', 'install'])]


@patch('plinth.modules.upgrades.distupgrade._apt_run')
def test_apt_autoremove(apt_run):
    """Test that apt autoremove works."""
    distupgrade._apt_autoremove()
    apt_run.assert_called_with(['autoremove'])


@patch('plinth.modules.upgrades.distupgrade._apt_run')
def test_apt_full_upgrade(apt_run):
    """Test that apt full upgrade works."""
    apt_run.return_value = 0
    distupgrade._apt_full_upgrade()
    apt_run.assert_called_with(
        ['full-upgrade', '-o', 'Dpkg::Options::=--force-confnew'])


@patch('subprocess.run')
def test_unatteneded_upgrades_run(run):
    """Test that running unattended upgrades works."""
    distupgrade._unattended_upgrades_run()
    run.assert_called_with(['unattended-upgrade', '--verbose'],
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                           check=False)


@patch('plinth.action_utils.service_restart')
def test_freedombox_restart(service_restart):
    """Test that restarting freedombox service works."""
    distupgrade._freedombox_restart()
    service_restart.assert_called_with('plinth')


@patch('subprocess.run')
def test_trigger_on_complete(run):
    """Test triggering post completion process."""
    distupgrade._trigger_on_complete()
    run.assert_called_with([
        'systemd-run', '--unit=freedombox-dist-upgrade-on-complete',
        '--description=Finish up upgrade to new stable Debian release',
        '/usr/bin/freedombox-cmd', 'upgrades', 'dist_upgrade_on_complete',
        '--no-args'
    ], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)


def test_on_complete(tmp_path):
    """Test that /etc/apt/sources.list is committed."""
    sources_list = tmp_path / 'sources.list'
    sources_list.write_text('before')
    temp_sources_list = tmp_path / 'sources.list.fbx-dist-upgrade'
    temp_sources_list.write_text('after')

    module = 'plinth.modules.upgrades.distupgrade'
    with patch(f'{module}.sources_list', sources_list), \
         patch(f'{module}.temp_sources_list', temp_sources_list):
        distupgrade.on_complete()
        assert sources_list.read_text() == 'after'
        assert not temp_sources_list.exists()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Application manifest for Home Assistant."""

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

_android_package_id = 'io.homeassistant.companion.android'

clients = [
    {
        'name': _('Home Assistant'),
        'platforms': [{
            'type': 'web',
            'url': '#'  # Filled in later
        }]
    },
    {
        'name':
            _('Home Assistant'),
        'platforms': [{
            'type': 'download',
            'os': 'macos',
            'url': 'https://apps.apple.com/us/app/home-assistant/id1099568401'
        }, {
            'type': 'store',
            'os': 'ios',
            'store_name': 'app-store',
            'url': 'https://apps.apple.com/us/app/home-assistant/id1099568401'
                   '?platform=iphone'
        }, {
            'type':
                'store',
            'os':
                'android',
            'store_name':
                'google-play',
            'url':
                store_url('google-play', 'io.homeassistant.companion.android')
        }, {
            'type':
                'store',
            'os':
                'android',
            'store_name':
                'f-droid',
            'url':
                store_url('f-droid',
                          'io.homeassistant.companion.android.minimal')
        }]
    }
]

backup = {
    'data': {
        'directories': ['/var/lib/home-assistant-freedombox/']
    },
    'services': ['home-assistant-freedombox']
}

tags = [
    _('Home Automation'),
    _('IoT'),
    _('Wi-Fi'),
    _('ZigBee'),
    _('Z-Wave'),
    _('Thread')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the upgrades module."""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^sys/upgrades/$', views.UpgradesConfigurationView.as_view(),
            name='index'),
    re_path(r'^sys/upgrades/activate-backports/$', views.activate_backports,
            name='activate-backports'),
    re_path(r'^sys/upgrades/firstboot/backports/$',
            views.BackportsFirstbootView.as_view(),
            name='backports-firstboot'),
    re_path(r'^sys/upgrades/upgrade/$', views.upgrade, name='upgrade'),
    re_path(r'^sys/upgrades/dist-upgrade/$', views.DistUpgradeView.as_view(),
            name='dist-upgrade'),
    re_path(r'^sys/upgrades/dist-upgrade/confirm/$',
            views.DistUpgradeConfirmView.as_view(),
            name='dist-upgrade-confirm'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
pytest configuration for all tests.
"""

import importlib
import os
import pathlib
from unittest.mock import patch

import pytest

try:
    importlib.import_module('splinter')
    importlib.import_module('selenium')
    _functional_libs_available = True
except ImportError:
    _functional_libs_available = False


def pytest_ignore_collect(collection_path, config):
    """Ignore functional tests when splinter is not available."""
    if collection_path.name == 'test_functional.py':
        return not _functional_libs_available


def pytest_configure(config):
    """Register additional markers, one for each app."""
    for app in (pathlib.Path(__file__).parent / 'modules').iterdir():
        if not app.is_dir():
            continue

        config.addinivalue_line('markers', app.name)


def pytest_collection_modifyitems(config, items):
    """Filter out specificly marked tests unless explicitly requested.

    The EXTENDED_TESTING environment variable is borrowed from the Lancaster
    consensus met by the Pearl community. See
    https://github.com/Perl-Toolchain-Gang/toolchain-site/blob/master/lancaster-consensus.md
    """

    def skip(item, reason):
        item.add_marker(pytest.mark.skip(reason=reason))

    extended = 'EXTENDED_TESTING' in os.environ
    if not (extended or config.getoption('--include-functional')):
        for item in items:
            if 'functional' in item.keywords or (
                    item.parent.fspath.basename
                    and item.parent.fspath.basename == 'test_functional.py'):
                skip(item, '--include-functional not provided')

    if not extended:
        for item in items:
            if 'heavy' in item.keywords:
                skip(item, ('Takes too much time. '
                            'Set EXTENDED_TESTING=1 to force run'))


@pytest.fixture(name='load_cfg')
def fixture_load_cfg():
    """Load test configuration."""
    from plinth import cfg

    keys = ('file_root', 'data_dir', 'custom_static_dir', 'store_file',
            'doc_dir', 'server_dir', 'host', 'port', 'use_x_forwarded_for',
            'use_x_forwarded_host', 'secure_proxy_ssl_header', 'box_name',
            'develop')
    saved_state = {}
    for key in keys:
        saved_state[key] = getattr(cfg, key)

    root_dir = pathlib.Path(__file__).resolve().parent
    cfg_file = root_dir / 'plinth' / 'develop.config'
    cfg.read_file(str(cfg_file))
    yield cfg

    for key in keys:
        setattr(cfg, key, saved_state[key])


@pytest.fixture(name='develop_mode')
def fixture_develop_mode(load_cfg):
    """Turn on development mode for a test."""
    load_cfg.develop = True
    yield
    load_cfg.develop = False


@pytest.fixture(name='needs_root', scope='session')
def fixture_needs_root():
    """Skip test if not running in root mode."""
    if os.geteuid() != 0:
        pytest.skip('Needs to be root')


@pytest.fixture(name='needs_not_root', scope='session')
def fixture_needs_not_root():
    """Skip test if running in root mode."""
    if os.geteuid() == 0:
        pytest.skip('Needs not to be root')


@pytest.fixture(name='needs_sudo')
def fixture_needs_sudo():
    """Skip test if sudo command is not available."""
    if not os.path.isfile('/usr/bin/sudo'):
        pytest.skip('Needs sudo command installed.')


@pytest.fixture(scope='session')
def splinter_selenium_implicit_wait():
    """Disable implicit waiting."""
    return 0


@pytest.fixture(scope='session')
def splinter_wait_time():
    """Disable explicit waiting."""
    return 0.01


@pytest.fixture(scope='session')
def splinter_driver_kwargs():
    """Disable smooth scrolling to fix 'Element x not clickable' errors."""
    from selenium.webdriver.firefox.options import Options
    driver_options = Options()
    driver_options.set_preference('general.smoothScroll', False)
    return {'options': driver_options}


@pytest.fixture(scope='session')
def splinter_browser_load_condition():
    """When a page it loaded, wait until <body> is available."""

    def _load_condition(browser):
        if browser.url == 'about:blank':
            return True

        ready_state = browser.execute_script('return document.readyState;')
        return ready_state == 'complete'

    return _load_condition


@pytest.fixture(name='mock_privileged')
def fixture_mock_privileged(request):
    """Mock the privileged decorator to nullify its effects."""
    try:
        privileged_modules_to_mock = request.module.privileged_modules_to_mock
    except AttributeError:
        raise AttributeError(
            'mock_privileged fixture requires "privileged_modules_to_mock" '
            'attribute at module level')

    mocked_methods = []
    for module_name in privileged_modules_to_mock:
        module = importlib.import_module(module_name)
        for name, member in module.__dict__.items():
            wrapped = getattr(member, '__wrapped__', None)
            if not callable(member) or not wrapped:
                continue

            if not getattr(member, '_privileged', False):
                continue

            while getattr(wrapped, '__wrapped__', None) and getattr(
                    wrapped, '_privileged', True):
                wrapped = getattr(wrapped, '__wrapped__', None)

            setattr(wrapped, '_skip_privileged_call', True)
            mocked_methods.append(wrapped)

    yield

    for mocked_method in mocked_methods:
        try:
            delattr(mocked_method, '_skip_privileged_call')
        except AttributeError:
            pass


@pytest.fixture(name='mock_run_as_user')
def fixture_mock_run_as_user():
    """A fixture to override action_utils.run_as_user."""

    def _bypass_runuser(*args, username, **kwargs):
        from plinth import action_utils
        return action_utils.run(*args, **kwargs)

    with patch('plinth.action_utils.run_as_user') as mock:
        mock.side_effect = _bypass_runuser
        yield


@pytest.fixture(name='splinter_screenshot_dir', scope='session')
def fixture_splinter_screenshot_dir(request):
    """Set default screenshot directory to ./screenshots.

    This can be overridden using --splinter-screenshot-dir=foo as the option.
    """
    option = request.config.getoption('--splinter-screenshot-dir')
    screenshots_dir = option if option != '.' else './screenshots'
    return os.path.abspath(screenshots_dir)


@pytest.fixture(autouse=True)
def fixture_fix_session_browser_screenshots(request):
    """Fix a bug in pytest-splinter for screenshots.

    When using session_browser, pytest-splinter does not take a screenshot when
    a test has failed. It is uses internal pytest API on the FixtureRequest
    object. This API was removed in later versions of pytest causing the
    failure. Re-implement the fixture that has the problem fixing this issue.

    Drop this fixture after a fix is merged and released in pytest-splinter.
    See: https://github.com/pytest-dev/pytest-splinter/pull/157
    """
    yield

    if not request.config.pluginmanager.has_plugin('pytest-splinter'):
        return

    session_tmpdir = request.getfixturevalue('session_tmpdir')
    splinter_session_scoped_browser = request.getfixturevalue(
        'splinter_session_scoped_browser')
    splinter_make_screenshot_on_failure = request.getfixturevalue(
        'splinter_make_screenshot_on_failure')
    splinter_screenshot_dir = request.getfixturevalue(
        'splinter_screenshot_dir')
    splinter_screenshot_getter_html = request.getfixturevalue(
        'splinter_screenshot_getter_html')
    splinter_screenshot_getter_png = request.getfixturevalue(
        'splinter_screenshot_getter_png')
    splinter_screenshot_encoding = request.getfixturevalue(
        'splinter_screenshot_encoding')

    # Screenshot for function scoped browsers is handled in
    # browser_instance_getter
    if not splinter_session_scoped_browser:
        return

    for name in request.fixturenames:
        fixture_def = request._fixture_defs.get(name)
        if not fixture_def or not fixture_def.cached_result:
            continue

        value = fixture_def.cached_result[0]
        should_take_screenshot = (hasattr(value, '__splinter_browser__')
                                  and splinter_make_screenshot_on_failure
                                  and getattr(request.node, 'splinter_failure',
                                              True))

        from pytest_splinter import plugin
        if should_take_screenshot:
            kwargs = {
                'request':
                    request,
                'fixture_name':
                    name,
                'session_tmpdir':
                    session_tmpdir,
                'browser_instance':
                    value,
                'splinter_screenshot_dir':
                    splinter_screenshot_dir,
                'splinter_screenshot_getter_html':
                    splinter_screenshot_getter_html,
                'splinter_screenshot_getter_png':
                    splinter_screenshot_getter_png,
                'splinter_screenshot_encoding':
                    splinter_screenshot_encoding
            }

            plugin._take_screenshot(**kwargs)


@pytest.fixture(name='host_sudo')
def fixture_host_sudo(host):
    """Pytest fixture to run commands with sudo."""
    with host.sudo():
        yield host


@pytest.fixture(name='test_menu')
def fixture_test_menu():
    """Initialized menu module."""
    from plinth import menu as menu_module

    menu_module.Menu._all_menus = set()
    menu_module.init()
    menu_module.Menu('home-id', name='Home', url_name='index')
    menu_module.Menu('apps-id', name='Apps', url_name='apps',
                     parent_url_name='index')
    menu_module.Menu('testapp-id', name='Test App', url_name='testapp:index',
                     parent_url_name='apps')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Configuration parser and default values for configuration options.
"""

import configparser
import logging
import os
import pathlib

logger = logging.getLogger(__name__)

# [Path] section
file_root = '/usr/share/plinth'
data_dir = '/var/lib/plinth'
custom_static_dir = '/var/www/plinth/custom/static'
store_file = data_dir + '/plinth.sqlite3'
doc_dir = '/usr/share/freedombox'
server_dir = '/freedombox'

# [Network] section
host = '127.0.0.1'
port = 8000

# Enable the following only if Plinth is behind a proxy server.  The
# proxy server should properly clean and the following HTTP headers:
#   X-Forwarded-For
#   X-Forwarded-Host
#   X-Forwarded-Proto
# If you enable these unnecessarily, this will lead to serious security
# problems. For more information, see
# https://docs.djangoproject.com/en/1.7/ref/settings/
#
# These are enabled by default in FreedomBox because the default
# configuration allows only connections from localhost
#
# Leave the values blank to disable
use_x_forwarded_for = True
use_x_forwarded_host = True
secure_proxy_ssl_header = 'HTTP_X_FORWARDED_PROTO'

# [Misc] section
box_name = 'FreedomBox'

# Other globals
develop = False

config_files = []


def expand_to_dot_d_paths(file_paths):
    """Expand a list of file paths to include file.d/* also."""
    final_list = []
    for file_path in file_paths:
        final_list.append(str(file_path))
        path = pathlib.Path(file_path)
        path_d = path.with_suffix(path.suffix + '.d')
        try:
            for dot_d_file in sorted(path_d.glob('*' + path.suffix)):
                final_list.append(str(dot_d_file))
        except Exception as exception:
            logger.warning('Unable to read from directory %s: %s', path_d,
                           exception)

    return final_list


def get_develop_config_path():
    """Return config path of current source folder for development mode."""
    root_directory = os.path.dirname(os.path.realpath(__file__))
    root_directory = os.path.realpath(root_directory)
    config_path = os.path.join(root_directory, 'develop.config')
    return config_path


def get_config_paths():
    """Get default config paths."""
    return [
        '/usr/share/freedombox/freedombox.config',
        '/etc/plinth/plinth.config',
        '/etc/freedombox/freedombox.config',
    ]


def read():
    """Read all configuration files."""
    config_paths = get_config_paths()
    for config_path in expand_to_dot_d_paths(config_paths):
        read_file(config_path)


def read_file(config_path):
    """Read and merge into defaults a single configuration file."""
    if not os.path.isfile(config_path):  # Does not throw exceptions
        # Ignore missing configuration files
        return

    # Keep a note of configuration files read.
    config_files.append(config_path)

    parser = configparser.ConfigParser(
        defaults={
            'parent_dir':
                pathlib.Path(config_path).parent.resolve(),
            'parent_parent_dir':
                pathlib.Path(config_path).parent.parent.resolve(),
        })
    parser.read(config_path)  # Ignores all read errors

    config_items = (
        ('Path', 'file_root', 'string'),
        ('Path', 'data_dir', 'string'),
        ('Path', 'custom_static_dir', 'string'),
        ('Path', 'store_file', 'string'),
        ('Path', 'doc_dir', 'string'),
        ('Path', 'server_dir', 'string'),
        ('Network', 'host', 'string'),
        ('Network', 'port', 'int'),
        ('Network', 'secure_proxy_ssl_header', 'string'),
        ('Network', 'use_x_forwarded_for', 'bool'),
        ('Network', 'use_x_forwarded_host', 'bool'),
        ('Misc', 'box_name', 'string'),
    )

    for section, name, datatype in config_items:
        try:
            value = parser.get(section, name)
        except (configparser.NoSectionError, configparser.NoOptionError):
            # Use default values for any missing keys in configuration
            continue
        else:
            if datatype == 'int':
                value = int(value)
            elif datatype == 'bool':
                value = (value.lower() == 'true')

            globals()[name] = value



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
TLS certificate configuration for postfix and dovecot.

See: https://ssl-config.mozilla.org/
See: http://www.postfix.org/TLS_README.html
See: https://doc.dovecot.org/configuration_manual/dovecot_ssl_configuration/
"""

import pathlib

from .. import postfix
from ..dovecot import is_version_24

# Mozilla Guideline v5.6, Postfix 1.17.7, OpenSSL 1.1.1d, intermediate
# Generated 2021-08
# https://ssl-config.mozilla.org/
_tls_medium_cipherlist = [
    'ECDHE-ECDSA-AES128-GCM-SHA256', 'ECDHE-RSA-AES128-GCM-SHA256',
    'ECDHE-ECDSA-AES256-GCM-SHA384', 'ECDHE-RSA-AES256-GCM-SHA384',
    'ECDHE-ECDSA-CHACHA20-POLY1305', 'ECDHE-RSA-CHACHA20-POLY1305',
    'DHE-RSA-AES128-GCM-SHA256', 'DHE-RSA-AES256-GCM-SHA384'
]

_postfix_config = {
    # Allow unencrypted auth on port 25, needed by Roundcube
    'smtpd_tls_auth_only': 'no',

    # Mozilla Intermediate Configuration
    'smtpd_tls_security_level': 'may',
    'smtpd_tls_mandatory_protocols': '!SSLv2, !SSLv3, !TLSv1, !TLSv1.1',
    'smtpd_tls_protocols': '!SSLv2, !SSLv3, !TLSv1, !TLSv1.1',
    'smtpd_tls_mandatory_ciphers': 'medium',
    'smtpd_tls_ciphers': 'medium',
    'tls_medium_cipherlist': ':'.join(_tls_medium_cipherlist),
    'tls_preempt_cipherlist': 'no',

    # Postfix SMTP client
    'smtp_tls_mandatory_protocols': '!SSLv2, !SSLv3, !TLSv1, !TLSv1.1',
    'smtp_tls_protocols': '!SSLv2, !SSLv3, !TLSv1, !TLSv1.1',
    'smtp_tls_mandatory_ciphers': 'medium',
    'smtp_tls_ciphers': 'medium',

    # Use DNSSEC to validate TLS certificates
    'smtp_host_lookup': 'dns',
    'smtp_dns_support_level': 'dnssec',
    'smtp_tls_security_level': 'dane',  # Opportunistic DANE TLS
}


def set_postfix_config(primary_domain, all_domains):
    """Set postfix configuration for TLS certificates."""
    items = []
    for domain in all_domains:
        items.append(f'{domain}=/etc/postfix/letsencrypt/{domain}/chain.pem')

    domain_map = ','.join(items)
    config = dict(_postfix_config)
    config.update({
        'tls_server_sni_maps':
            f'inline:{{{domain_map}}}',
        'smtpd_tls_chain_files':
            f'/etc/postfix/letsencrypt/{primary_domain}/chain.pem',
        'smtpd_tls_cert_file':
            ''
    })
    postfix.set_config(config)


def set_dovecot_config(primary_domain, all_domains):
    """Set dovecot configuration for TLS certificates."""
    is_new_version = is_version_24()

    # Determine whether to prefix file paths with '<' based on version
    prefix = ''
    cert_naming = 'ssl_server_cert_file'
    key_naming = 'ssl_server_key_file'
    if not is_new_version:
        prefix = '<'
        cert_naming = 'ssl_cert'
        key_naming = 'ssl_key'

    content = f'''# This file is managed by FreedomBox
{cert_naming} = {prefix}/etc/dovecot/letsencrypt/{primary_domain}/cert.pem
{key_naming} = {prefix}/etc/dovecot/letsencrypt/{primary_domain}/privkey.pem
'''

    for domain in all_domains:
        content += f'''
local_name {domain} {{
  {cert_naming} = {prefix}/etc/dovecot/letsencrypt/{domain}/cert.pem
  {key_naming} = {prefix}/etc/dovecot/letsencrypt/{domain}/privkey.pem
}}
'''
    cert_config = pathlib.Path('/etc/dovecot/conf.d/91-freedombox-tls.conf')
    cert_config.write_text(content, encoding='utf-8')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Names App."""

import pathlib

import augeas

from plinth import action_utils
from plinth.actions import privileged

fallback_conf = pathlib.Path(
    '/etc/systemd/resolved.conf.d/freedombox-fallback.conf')
override_conf = pathlib.Path('/etc/systemd/resolved.conf.d/freedombox.conf')
source_fallback_conf = pathlib.Path(
    '/usr/share/freedombox'
    '/etc/systemd/resolved.conf.d/freedombox-fallback.conf')

HOSTS_LOCAL_IP = '127.0.1.1'


@privileged
def set_hostname(hostname: str):
    """Set system hostname using hostnamectl."""
    action_utils.run(
        ['hostnamectl', 'set-hostname', '--transient', '--static', hostname],
        check=True)
    action_utils.service_restart('avahi-daemon')


def _load_augeas_hosts():
    """Initialize Augeas for editing /etc/hosts."""
    hosts_file = '/etc/hosts'
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('hosts', hosts_file)
    aug.set('/augeas/context', '/files' + hosts_file)
    aug.load()
    return aug


def get_old_domains(aug=None) -> list[str]:
    """Return the list of domains store in old /etc/hosts format."""
    if not aug:
        aug = _load_augeas_hosts()

    domains = {}  # Maintain order of entries
    for match in aug.match('*'):
        if aug.get(match + '/ipaddr') != HOSTS_LOCAL_IP:
            continue

        domain = aug.get(match + '/canonical')
        aliases = []
        for alias_match in aug.match(match + '/alias'):
            aliases.append(aug.get(alias_match))

        # Read old style domains too.
        if aliases and '.' not in aliases[-1]:
            hostname = aliases[-1]
            if domain.startswith(hostname + '.'):
                domain = domain.partition('.')[2]
                aliases = aliases[:-1]

        for value in [domain] + aliases:
            if value:
                domains[domain] = True

    return list(domains.keys())


@privileged
def domain_delete_all():
    """Remove all static domain names from /etc/hosts."""
    aug = _load_augeas_hosts()
    for match in aug.match('*'):
        if aug.get(match + '/ipaddr') == HOSTS_LOCAL_IP:
            aug.remove(match)

    aug.save()


@privileged
def install_resolved():
    """Install systemd-resolved related packages."""
    packages = ['systemd-resolved', 'libnss-resolve']
    action_utils.run(['dpkg', '--configure', '-a'], check=False)
    with action_utils.apt_hold_freedombox():
        action_utils.run_apt_command(['--fix-broken', 'install'])
        returncode = action_utils.run_apt_command(['install'] + packages)

    if returncode:
        raise RuntimeError(
            f'Apt command failed with return code: {returncode}')


@privileged
def set_resolved_configuration(dns_fallback: bool | None = None,
                               dns_over_tls: str | None = None,
                               dnssec: str | None = None):
    """Set systemd-resolved configuration options."""
    if dns_fallback is not None:
        _set_enable_dns_fallback(dns_fallback)

    if dns_over_tls is not None or dnssec is not None:
        _set_resolved_configuration(dns_over_tls, dnssec)

    # Workaround buggy reload that does not apply DNS-over-TLS changes
    # properly.
    action_utils.service_try_restart('systemd-resolved')


def get_resolved_configuration() -> dict[str, bool]:
    """Return systemd-resolved configuration."""
    configuration = _get_resolved_configuration()
    configuration['dns_fallback'] = fallback_conf.exists()
    return configuration


def _set_enable_dns_fallback(dns_fallback: bool):
    """Update whether to use DNS fallback servers."""
    if dns_fallback:
        if not fallback_conf.exists():
            fallback_conf.parent.mkdir(parents=True, exist_ok=True)
            fallback_conf.symlink_to(source_fallback_conf)
    else:
        fallback_conf.unlink(missing_ok=True)


def _load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('Systemd', str(override_conf))
    aug.set('/augeas/context', '/files' + str(override_conf))
    aug.load()
    return aug


def _get_resolved_configuration():
    """Return overridden configuration for systemd-resolved."""
    aug = _load_augeas()
    # Default value for DNSSEC upstream is 'allow-downgrade', but in Debian it
    # is 'no'.
    return {
        'dns_over_tls': aug.get('Resolve/DNSOverTLS/value') or 'no',
        'dnssec': aug.get('Resolve/DNSSEC/value') or 'no'
    }


def _set_resolved_configuration(dns_over_tls: str | None = None,
                                dnssec: str | None = None):
    """Write configuration into a systemd-resolved override file."""
    aug = _load_augeas()

    if dns_over_tls is not None:
        aug.set('Resolve/DNSOverTLS/value', dns_over_tls)

    if dnssec is not None:
        aug.set('Resolve/DNSSEC/value', dnssec)

    aug.save()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Test module for Operation and OperationsManager."""

import threading
import time
from collections import OrderedDict
from unittest.mock import Mock, call, patch

import pytest

from plinth import app

from .. import operation as operation_module
from ..notification import Notification
from ..operation import Operation, OperationsManager


class AppTest(app.App):
    app_id = 'testapp'

    def __init__(self):
        super().__init__()

        info = app.Info(self.app_id, 1, name='Test App')
        self.add(info)


@patch('plinth.operation.Operation._update_notification')
def test_operation_default_initialization(update_notification):
    """Test Operation initialization with default values."""
    target = Mock()
    operation = Operation('testid', 'testapp', 'op1', target)
    assert operation.op_id == 'testid'
    assert operation.app_id == 'testapp'
    assert operation.name == 'op1'
    assert operation.show_message
    assert not operation.show_notification
    assert operation.target == target
    assert operation.args == []
    assert operation.kwargs == {}
    assert operation.on_complete is None
    assert operation.state == Operation.State.WAITING
    assert operation._message is None
    assert operation.exception is None
    assert operation.thread_data == {}
    assert isinstance(operation.thread, threading.Thread)
    assert operation.thread._operation == operation
    update_notification.assert_has_calls([call()])


@patch('plinth.operation.Operation._update_notification')
def test_operation_initialization(update_notification):
    """Test Operation initialization with explicit values."""
    on_complete = Mock()
    operation = Operation('testid', 'testapp', 'op1', Mock(), ['arg1'],
                          {'arg2': 'value2'}, False, True,
                          {'data1': 'datavalue1'}, on_complete)
    assert not operation.show_message
    assert operation.show_notification
    assert operation.args == ['arg1']
    assert operation.kwargs == {'arg2': 'value2'}
    assert operation.on_complete == on_complete
    assert operation.state == Operation.State.WAITING
    assert operation._message is None
    assert operation.exception is None
    assert operation.thread_data == {'data1': 'datavalue1'}
    update_notification.assert_has_calls([call()])


def test_operation_str():
    """Test string representation of operation."""
    operation = Operation('testid', 'testapp', 'op1', Mock())
    assert str(operation) == 'Operation: testapp: op1'


@patch('plinth.operation.Operation._update_notification')
def test_successful_operation(update_notification):
    """Test running a operation that succeeds."""
    target = Mock()
    target.return_value = 'test-return'
    on_complete = Mock()
    operation = Operation('testid', 'testapp', 'op1', target, ['arg1'],
                          {'arg2': 'value2'}, on_complete=on_complete)
    operation.run()
    assert operation.join() == 'test-return'
    target.assert_has_calls([call('arg1', arg2='value2')])
    assert operation.state == Operation.State.COMPLETED
    assert operation.return_value == 'test-return'
    on_complete.assert_has_calls([call(operation)])
    update_notification.assert_has_calls([call(), call()])


@patch('plinth.operation.Operation._update_notification')
def test_error_operation(update_notification):
    """Test running an operation that fails."""
    target = Mock()
    target.side_effect = RuntimeError('error1')
    on_complete = Mock()
    operation = Operation('testid', 'testapp', 'op1', target, ['arg1'],
                          {'arg2': 'value2'}, on_complete=on_complete)
    operation.run()
    with pytest.raises(RuntimeError):
        operation.join()

    target.assert_has_calls([call('arg1', arg2='value2')])
    assert operation.state == Operation.State.COMPLETED
    assert operation.exception == target.side_effect
    on_complete.assert_has_calls([call(operation)])
    update_notification.assert_has_calls([call(), call()])


@patch('plinth.operation.Operation._update_notification')
def test_join_before_start(update_notification):
    """Test waiting until operation finishes.."""
    event = threading.Event()
    operation = Operation('testid', 'testapp', 'op1', Mock)
    success = []

    def _wait():
        """Wait for operation to start."""
        event.set()
        operation.join()
        success.append(True)

    thread = threading.Thread(target=_wait)
    thread.start()
    event.wait()
    time.sleep(0.1)  # Ensure that thread is waiting before we start operation.
    operation.run()
    thread.join()
    assert success


@patch('plinth.operation.Operation._update_notification')
def test_join_raises_exception(update_notification):
    """Test that joining raises exception if thread does.."""
    target = Mock()
    target.side_effect = RuntimeError('error1')
    on_complete = Mock()
    operation = Operation('testid', 'testapp', 'op1', target, ['arg1'],
                          {'arg2': 'value2'}, on_complete=on_complete)
    operation.run()
    with pytest.raises(RuntimeError):
        operation.join()


def test_getting_operation_from_thread():
    """Test that operation object can be retread from within the thread."""

    def target():
        operation = Operation.get_operation()
        operation.thread_data['test_operation'] = operation

    operation = Operation('testid', 'testapp', 'op1', target)
    operation.run()
    operation.join()
    assert operation.thread_data['test_operation'] == operation


@patch('plinth.operation.Operation._update_notification')
def test_updating_operation(update_notification):
    """Test that operation object can be updated from within the thread."""
    exception = RuntimeError('error1')

    def target():
        operation = Operation.get_operation()
        operation.on_update('message1', exception)

    operation = Operation('testid', 'testapp', 'op1', target)
    operation.run()
    with pytest.raises(RuntimeError):
        operation.join()

    assert operation._message == 'message1'
    assert operation.exception == exception
    update_notification.assert_has_calls([call(), call(), call()])


@patch('plinth.app.App.get')
def test_message(app_get):
    """Test getting the operation's message."""
    operation = Operation('testid', 'testapp', 'op1', Mock())
    operation._message = 'message1'
    operation.exception = RuntimeError('error1')
    assert operation.message == 'message1'
    assert operation.translated_message == 'message1'

    operation._message = None
    assert operation.message == 'Error: {name}: {exception}'
    assert operation.translated_message == 'Error: op1: error1'

    operation.exception = None
    operation.state = Operation.State.WAITING
    assert operation.message == 'Waiting to start: {name}'
    assert operation.translated_message == 'Waiting to start: op1'

    operation.exception = None
    operation.state = Operation.State.RUNNING
    assert operation.message == '{name}'
    assert operation.translated_message == 'op1'

    operation.exception = None
    operation.state = Operation.State.COMPLETED
    assert operation.message == 'Finished: {name}'
    assert operation.translated_message == 'Finished: op1'


@patch('plinth.app.App.get')
@pytest.mark.django_db
def test_update_notification(app_get):
    """Test that operation notification is created."""
    app_get.return_value = AppTest()
    operation = Operation('testid', 'testapp', 'op1', Mock(),
                          show_notification=True)
    note = Notification.get('testapp-operation')
    assert note.id == 'testapp-operation'
    assert note.app_id == 'testapp'
    assert note.severity == 'info'
    assert note.title == 'Test App'
    assert note.message == operation.message
    assert note.body_template == 'operation-notification.html'
    assert note.group == 'admin'
    assert not note.dismissed
    assert note.data['app_name'] == 'Test App'
    assert note.data['app_icon'] is None
    assert note.data['app_icon_filename'] is None
    assert note.data['state'] == 'waiting'
    assert note.data['exception'] is None
    assert note.data['name'] == 'translate:op1'

    operation.exception = RuntimeError()
    operation._update_notification()
    note = Notification.get('testapp-operation')
    assert note.severity == 'error'


def test_manager_global_instance():
    """Test that single global instance of operation's manager is available."""
    assert isinstance(operation_module.manager, OperationsManager)


def test_manager_init():
    """Test initializing operations manager."""
    manager = OperationsManager()
    assert manager._operations == {}
    assert manager._current_operation is None
    assert isinstance(manager._lock, threading.RLock().__class__)


def test_manager_new():
    """Test creating a new operation using a manager."""
    manager = OperationsManager()
    event = threading.Event()

    def target():
        event.wait()

    operation = manager.new('testop', 'testapp', 'op1', target)
    assert isinstance(operation, Operation)
    assert manager._current_operation == operation
    assert manager._operations == {'testop': operation}

    event.set()
    operation.join()
    assert manager._current_operation is None
    assert manager._operations == OrderedDict(testop=operation)


def test_manager_new_without_show_message():
    """Test creating an operation that does not show message."""
    manager = OperationsManager()
    event = threading.Event()

    def target():
        event.wait()

    operation = manager.new('testop', 'testapp', 'op1', target,
                            show_message=False)
    event.set()
    operation.join()
    assert manager._current_operation is None
    assert manager._operations == {}


def test_manager_new_raises():
    """Test that a new operation is always unique."""
    manager = OperationsManager()
    event1 = threading.Event()
    event2 = threading.Event()

    operation1 = manager.new('testop1', 'testapp', 'op1',
                             lambda: event1.wait())

    # Creating operation with different ID works
    operation2 = manager.new('testop2', 'testapp', 'op3', lambda: event2.set())

    assert manager._operations == OrderedDict(testop1=operation1,
                                              testop2=operation2)

    # Creating operation with same id while the operation is running/scheduled
    # throws exception
    with pytest.raises(KeyError):
        manager.new('testop1', 'testapp', 'op1', Mock())

    # Creating operation with same id after the operation is completed works.
    event1.set()  # Allow the first operation to complete
    event2.wait(10)  # Wait until the second operation has started
    operation1_new = manager.new('testop1', 'testapp', 'op1', Mock())

    assert manager._operations == OrderedDict(testop1=operation1_new,
                                              testop2=operation2)


def test_manager_scheduling():
    """Test creating a multiple operations and scheduling them."""
    manager = OperationsManager()
    event1 = threading.Event()
    event2 = threading.Event()
    event3 = threading.Event()

    operation1 = manager.new('testop1', 'testapp', 'op1', event1.wait)
    operation2 = manager.new('testop2', 'testapp', 'op2', event2.wait)
    operation3 = manager.new('testop3', 'testapp', 'op3', event3.wait)

    def _assert_is_running(current_operation):
        assert manager._current_operation == current_operation
        assert manager._operations == OrderedDict(testop1=operation1,
                                                  testop2=operation2,
                                                  testop3=operation3)
        for operation in [operation1, operation2, operation3]:
            alive = (operation == current_operation)
            assert operation.thread.is_alive() == alive

    _assert_is_running(operation1)
    event1.set()
    operation1.join()

    _assert_is_running(operation2)
    event2.set()
    operation2.join()

    _assert_is_running(operation3)
    event3.set()
    operation3.join()


def test_manager_filter():
    """Test returning filtered operations."""
    manager = OperationsManager()
    operation1 = manager.new('testop1', 'testapp1', 'op1', Mock())
    operation2 = manager.new('testop2', 'testapp1', 'op2', Mock())
    operation3 = manager.new('testop3', 'testapp2', 'op3', Mock())
    manager.filter('testapp1') == [operation1, operation2]
    manager.filter('testapp2') == [operation3]


def test_manager_collect_results():
    """Test collecting results from the manager."""
    manager = OperationsManager()
    event = threading.Event()

    operation1 = manager.new('testop1', 'testapp1', 'op1', Mock())
    operation2 = manager.new('testop2', 'testapp2', 'op2', Mock())
    operation3 = manager.new('testop3', 'testapp1', 'op3', event.wait)
    operation1.join()
    operation2.join()
    assert manager.collect_results('testapp1') == [operation1]
    assert manager._operations == OrderedDict(testop2=operation2,
                                              testop3=operation3)
    event.set()
    operation3.join()
    assert manager.collect_results('testapp1') == [operation3]
    assert manager._operations == OrderedDict(testop2=operation2)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Forms for BIND module."""

from django import forms
from django.core.validators import validate_ipv46_address
from django.utils.translation import gettext_lazy as _


def validate_ips(ips):
    """Validate that ips is a list of IP addresses, separated by space."""
    for ip_addr in ips.split():
        validate_ipv46_address(ip_addr)


class BindForm(forms.Form):
    """BIND configuration form."""

    forwarders = forms.CharField(
        label=_('Forwarders'), required=False, validators=[validate_ips],
        help_text=_('A list DNS servers, separated by space, to which '
                    'requests will be forwarded'))
